from __future__ import annotations

import abc
import datetime
import gzip
import hashlib
import io
import tempfile
import typing as t
from pathlib import Path
from shutil import copyfileobj

from typing_extensions import Self

from datapane._vendor import base64io
from datapane.common import guess_type

SERVED_REPORT_ASSETS_DIR = "assets"
GZIP_MTIME = datetime.datetime(year=2000, month=1, day=1).timestamp()


class FileEntry:
    file: t.IO
    _ext: str
    _dir_path: t.Optional[Path]

    # post-freeze
    frozen: bool = False
    mime: str
    hash: str
    size: int
    wrapped: t.BinaryIO

    def __init__(self, ext: str, mime: t.Optional[str] = None, dir_path: t.Optional[Path] = None):
        self.mime = mime or guess_type(Path(f"tmp{ext}"))
        self._ext = ext
        self._dir_path = dir_path

    @abc.abstractmethod
    def freeze(self) -> None:
        """Must be called after writing / adding to store
        # TODO - add to contextmanager??
        """

    @property
    @abc.abstractmethod
    def src(self) -> str:
        pass

    def as_dict(self) -> dict:
        assert self.frozen
        return dict(src=self.src, hash=self.hash, size=self.size, mime=self.mime)

    def __eq__(self, other: FileEntry) -> bool:
        if self.hash:
            return self.hash == other.hash
        raise NotImplementedError()


class NullWriter(io.BytesIO):
    def write(self, s):
        pass

    def writelines(self, *a, **kw) -> None:
        pass


class DummyFileEntry(FileEntry):
    """File entry that discards all data - for internal use"""

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self.file = NullWriter()

    def freeze(self):
        self.size = 0
        self.hash = "abcdef"
        self.frozen = True

    def src(self) -> str:
        return "/dev/null"


class B64FileEntry(FileEntry):
    """Memory-based b64 file"""

    # requires b64io is bytes only and wraps to a bytes file only
    file: base64io.Base64IO
    wrapped: io.BytesIO
    contents: bytes

    def __init__(self, ext: str, mime: t.Optional[str] = None, *a, **kw):
        super().__init__(ext, mime, *a, **kw)
        self.wrapped = io.BytesIO()
        self.file = base64io.Base64IO(self.wrapped)

    def freeze(self) -> None:
        if not self.frozen:
            self.frozen = True
            # get a reference to the buffer to splice later
            self.file.close()
            self.file.flush()
            self.contents = self.wrapped.getvalue()
            # calc other properties
            self.hash = hashlib.sha256(self.contents).hexdigest()[:10]
            self.size = self.wrapped.tell()

    @property
    def src(self) -> str:
        return f"data:{self.mime};base64,{self.contents.decode('ascii')}"


class GzipTmpFileEntry(FileEntry):
    """Gzipped file, by default stored in /tmp"""

    # both file and wapper files are bytes-only
    file: gzip.GzipFile
    # TODO - this could actually be an in-memory file...
    wrapped: tempfile.NamedTemporaryFile
    has_output_dir: bool = False

    # Do we need DPTmpFile here, or just use namedtempfile??
    def __init__(self, ext: str, mime: t.Optional[str] = None, dir_path: t.Optional[Path] = None):
        super().__init__(ext, mime, dir_path)

        if dir_path:
            # create as a permanent file within the given dir
            self.has_output_dir = True
            self.wrapped = tempfile.NamedTemporaryFile("w+b", suffix=ext, prefix="dp-", dir=dir_path, delete=False)
        else:
            self.wrapped = tempfile.NamedTemporaryFile("w+b", suffix=ext, prefix="dp-")

        self.file = gzip.GzipFile(fileobj=self.wrapped, mode="w+b", mtime=GZIP_MTIME)

    def calc_hash(self, f: t.IO) -> str:
        f.seek(0)
        file_hash = hashlib.sha256()
        while chunk := f.read(8192):
            file_hash.update(chunk)
        return file_hash.hexdigest()[:10]

    @property
    def src(self) -> str:
        if self.has_output_dir:
            return f"/{SERVED_REPORT_ASSETS_DIR}/{Path(self.wrapped.name).name}"
        else:
            return "NYI"

    def freeze(self) -> None:
        if not self.frozen:
            self.frozen = True
            self.file.flush()
            self.file.close()
            self.wrapped.flush()
            # size will be the compressed size...
            self.size = self.wrapped.tell()
            self.hash = self.calc_hash(self.wrapped)


class FileStore:
    # TODO - make this a CAS (index by object hash itself?)
    # NOTE - currently we pass dir_path via the FileStore, could move into the file themselves?
    def __init__(self, fw_klass: t.Type[FileEntry], assets_dir: t.Optional[Path] = None):
        super().__init__()
        self.fw_klass = fw_klass
        self.files: t.List[FileEntry] = []
        self.dir_path = assets_dir

    def __add__(self, other: FileStore) -> Self:
        # TODO - ensure factory is the same for both
        self.files.extend(other.files)
        return self

    @property
    def store_count(self) -> int:
        return len(self.files)

    @property
    def file_list(self) -> t.List[t.BinaryIO]:
        return [f.wrapped for f in self.files]

    def get_file(self, ext: str, mime: str) -> FileEntry:
        return self.fw_klass(ext, mime, self.dir_path)

    def add_file(self, fw: FileEntry) -> None:
        fw.freeze()
        self.files.append(fw)

    def load_file(self, path: Path) -> FileEntry:
        """load a file into the store (makes a copy)"""
        # TODO - ideally lazily-link a path to the store (rather than copy it in)
        ext = "".join(path.suffixes)
        dest_obj = self.fw_klass(ext=ext, dir_path=self.dir_path)
        with path.open("rb") as src_obj:
            copyfileobj(src_obj, dest_obj.file)
        self.add_file(dest_obj)
        return dest_obj

    def as_dict(self) -> dict:
        """Build a json structure suitable for embedding in a html file, json-rpc response, etc."""
        x: FileEntry  # noqa: F842
        return {x.hash: x.as_dict() for x in self.files}

    def get_entry(self, hash: str) -> t.Optional[FileEntry]:
        # TODO - change self.files to a dict[hash, FileEntry]?
        return next((f for f in self.files if f.hash == hash), None)



# Sample Datapane App - {{ name }}

import altair as alt
import pandas as pd

import datapane as dp
import datapane.blocks.asset

# download data & group by manufacturer
df = pd.read_csv(
    "https://covid.ourworldindata.org/data/vaccinations/vaccinations-by-manufacturer.csv", parse_dates=["date"]
)
df = df.groupby(["vaccine", "date"])["total_vaccinations"].sum().tail(1000).reset_index()

# plot vaccinations over time using Altair
plot = (
    alt.Chart(df)
    .mark_area(opacity=0.4, stroke="black")
    .encode(
        x="date:T",
        y=alt.Y("total_vaccinations:Q"),
        color=alt.Color("vaccine:N", scale=alt.Scale(scheme="set1")),
        tooltip="vaccine:N",
    )
    .interactive()
    .properties(width="container")
)

# tablulate total vaccinations by manufacturer
total_df = df[df["date"] == df["date"].max()].sort_values("total_vaccinations", ascending=False).reset_index(drop=True)
total_styled = total_df.style.bar(
    subset=["total_vaccinations"], color="#5fba7d", vmax=total_df["total_vaccinations"].sum()
)

# embed into a Datapane Report
report = dp.App(
    "## Vaccination Report",
    datapane.blocks.asset.Plot(plot, caption="Vaccinations by manufacturer over time"),
    datapane.blocks.asset.Table(total_styled, caption="Current vaccination totals by manufacturer"),
    datapane.blocks.asset.Table(df, caption="Initial Dataset"),
)
report.upload(
    name="Covid Vaccinations Demo {{ name }}",
    description="Covid Vaccinations report, using data from ourworldindata",
    open=True,
)



import json
from collections import namedtuple
import os
import shutil
from typing import Dict, Optional, List
import warnings
import pandas as pd
from functools import lru_cache
from polly.errors import (
    InvalidSchemaJsonException,
    InvalidSyntaxForRequestException,
    EmptyPayloadException,
    RequestException,
    UnauthorizedException,
    extract_json_api_error,
    paramException,
)
from polly.auth import Polly
from polly.cohort import Cohort
from polly import helpers, constants as const, application_error_info as app_err_info
from polly.help import example
import polly.http_response_codes as http_codes
from polly.constants import SUPPORTED_ENTITY_TYPES, CURATION_COHORT_CACHE
from polly.helpers import get_cohort_constants
from polly.tracking import Track


class Curation:
    """
    The Curation class contains wrapper functions around the models used for
    semantic annotations of string/text.
    Args:
        token (str): token copy from polly.

    Usage:
            from polly.curation import Curation

            curationObj = Curation(token)
    """

    Tag = namedtuple("Tag", ["name", "ontology_id", "entity_type"])
    example = classmethod(example)

    def __init__(
        self,
        token=None,
        env="",
        default_env="polly",
    ) -> None:
        # check if COMPUTE_ENV_VARIABLE present or not
        # if COMPUTE_ENV_VARIABLE, give priority
        env = helpers.get_platform_value_from_env(
            const.COMPUTE_ENV_VARIABLE, default_env, env
        )
        self.session = Polly.get_session(token, env=env)
        self.base_url = f"https://v2.api.{self.session.env}.elucidata.io"
        self.discover_url = f"https://api.discover.{self.session.env}.elucidata.io"
        self.elastic_url = (
            f"https://api.datalake.discover.{self.session.env}.elucidata.io/elastic/v2"
        )
        self.inference_url = f"https://api.discover.{self.session.env}.elucidata.io/curations/inferences/"
        self.cohort = Cohort()
        self.cohort_constants = get_cohort_constants()

    def _handle_errors(self, response):
        detail = response.get("errors")[0].get("detail", [])
        title = response.get("errors")[0].get("title", [])
        return title, detail

    def _fetch_metadata_from_cohort(self, repo_name: str, dataset_ids: List[str]):
        """
        Utility function for fetching metadata using cohorts.

        Arguments:
            repo_name (str) : name of the repository for fetching datasets.
            dataset_ids (List[str]): dataset ids to be used for inference

        Returns:
            Returns sample metadata, dataset and sample ids.
        """
        sample_metadata = {}
        dataset_to_sample_id = {"dataset_id": [], "sample_id": []}

        if not (os.path.isdir(CURATION_COHORT_CACHE)):
            os.mkdir(CURATION_COHORT_CACHE)
        else:
            shutil.rmtree(CURATION_COHORT_CACHE)
            os.mkdir(CURATION_COHORT_CACHE)

        self.cohort.create_cohort(
            CURATION_COHORT_CACHE, "sample_metadata_query", "desc"
        )

        # Fetch metadata using cohorts
        for dataset_id in dataset_ids:
            datasets_sample_metadata = []

            if not (
                repo_name in self.cohort_constants
                and self.cohort_constants[repo_name]["file_structure"] != "multiple"
            ):
                # multiple mapped repo such as GEO
                self.cohort.add_to_cohort(repo_name, dataset_id=dataset_id)
            else:
                # for single mapped repos such as TCGA
                self.cohort.add_to_cohort(repo_name, dataset_id=[dataset_id])

            col_metadata = self.cohort.merge_data("sample")
            all_sample_ids = col_metadata.index.tolist()

            col_metadata.loc[:, "dataset_id"] = dataset_id
            dataset_to_sample_id["dataset_id"] += [dataset_id] * len(all_sample_ids)

            col_metadata.loc[:, "sample_id"] = all_sample_ids
            dataset_to_sample_id["sample_id"] += all_sample_ids

            datasets_sample_metadata += list(col_metadata.T.to_dict().values())

            if not (
                repo_name in self.cohort_constants
                and self.cohort_constants[repo_name]["file_structure"] != "multiple"
            ):
                self.cohort.remove_from_cohort(dataset_id)
            else:
                self.cohort.remove_from_cohort([dataset_id])

            sample_metadata[dataset_id] = datasets_sample_metadata

        dataset_to_sample_id = pd.DataFrame.from_dict(dataset_to_sample_id)

        return sample_metadata, dataset_to_sample_id

    def _clinical_model_param_checks(
        self,
        repo_name: str,
        dataset_ids: List[str],
        sample_ids: Optional[List[str]] = None,
    ):
        """
        Checking the parameter passed to the clinical label assigning model.

        Arguments:
            repo_name (str): repo name
            dataset_ids (list[str]): list of dataset ids

        Keyword Arguments:
            sample_ids (list[str], optional): Optional Parameter. List of sample ids.
            Default is 'None'.

        Raises:
            paramException
        """
        if dataset_ids is None or type(dataset_ids) is not list:
            raise paramException(
                title="Param Exception",
                detail="Dataset IDs should be given as a valid list of strings",
            )

        if sample_ids is not None and type(sample_ids) is not list:
            raise paramException(
                title="Param Exception",
                detail="Sample IDs should be given as a valid list of strings",
            )

        if repo_name != "geo" and not any(
            ["GSE" in dataset_id for dataset_id in dataset_ids]
        ):
            warnings.warn(
                "The model is tested with GEO metadata and the labels may be wrong for other repos"
            )

    def _post_process_clinical_tags(
        self,
        clinical_tags: pd.DataFrame,
        is_sample_tag: bool,
        sample_ids: Optional[List[str]] = None,
    ) -> pd.DataFrame:
        """
        process the response of the model (dataframe with clinical tags and samples)
        and return relevant feilds.
        incase no sample_ids are provided by the user, we return the dataset_ids and the clinical tags
        incase sample_ids are also provided, then we return the dataset_ids, the sample_ids and the clincal tags.

        Arguments:
            clinical_tags (pd.DataFrame): dataframe of the sample_ids and assigned clinical tags
            is_sample_tag (bool): if samples passed

        Keyword Arguments:
            sample_ids (list[str]): list of sample ids (default: {None})

        Returns:
            a dataframe with the the dataset_ids, sample_ids and the assigned clinical tags
        """
        if is_sample_tag:
            # if the user has provided list of samples, then we filter in just those sample ids
            # for the dataset ids.
            # taking only those clinical tags and samples where the sample_ids are in the sample_id list
            # provided by the user.
            clinical_tags = clinical_tags[
                clinical_tags["sample_id"].isin(sample_ids)
            ].reset_index(drop=True)

            # in case the sample_ids provided by the user are not present in the dataset_ids provided.
            if clinical_tags.empty or clinical_tags.shape[0] < len(sample_ids):
                warnings.warn(
                    "The output is empty or has missing sample ids because they are not present in given datasets."
                )

            # return sample level tags here
            return clinical_tags
        # if no sample_ids were passed by the user, then
        # returning dataset level tags by removing sample id and removing duplicate columns
        return (
            clinical_tags.drop(columns=["sample_id"])
            .drop_duplicates()
            .reset_index(drop=True)
        )

    def _handle_perform_inference_api_error(self, response):
        if response.status_code == http_codes.UNAUTHORIZED:
            raise UnauthorizedException("User is unauthorized to access this")
        elif response.status_code == http_codes.BAD_REQUEST:
            title, details = extract_json_api_error(response)
            if title == app_err_info.EMPTY_PAYLOAD_CODE:
                raise EmptyPayloadException()
            elif app_err_info.INVALID_MODEL_NAME_TITLE in title:
                raise InvalidSyntaxForRequestException()
        elif response.status_code == http_codes.INTERNAL_SERVER_ERROR:
            raise InvalidSchemaJsonException()
        elif response.status_code == http_codes.GATEWAY_TIMEOUT:
            error_detail = response.json().get("message", "Request timed out")
            raise Exception(error_detail)
        else:
            title, details = extract_json_api_error(response)
            raise Exception("Exception Occurred :" + str(details))

    def _perform_inference(
        self,
        model_name: str,
        input_data: dict,
    ) -> dict:
        """
        This is a wrapper around model inference APIs
        It serializes input_data, calls the API for the given model_name
        and returns deserialized output.

        Args:
            model_name (str): one of 'normalizer', 'biobert' and 'control-perturbation'
            input_data (dict): model input

        Returns:
            dict
        """
        url = self.inference_url + model_name

        payload = {}
        payload = json.dumps({"data": {"attributes": input_data, "type": "curation"}})
        response = self.session.post(url, data=payload)
        try:
            if response.status_code != 201:
                self._handle_perform_inference_api_error(response)
        except Exception as err:
            raise err
        try:
            response = response.json()
        except json.JSONDecodeError as e:
            raise e
        if "data" in response:
            return response.get("data")
        return response

    @Track.track_decorator
    @lru_cache(maxsize=None)
    def standardise_entity(
        self,
        mention: str,
        entity_type: str,
        context: Optional[str] = None,
        threshold: Optional[float] = None,
    ) -> dict:
        """
        Map a given mention (keyword) to an ontology term.
        Given a text and entity type, users can get the Polly compatible ontology for the text such as the MESH ontology.

        Args:
            mention (str): mention of an entity e.g. "Cadiac arrythmia"
            entity_type (str): Should be one of\
            ['disease', 'drug', 'tissue', 'cell_type', 'cell_line', 'species', 'gene']
            context (str): The text where the mention occurs.\
            This is used to resolve abbreviations.
            Threshold: (float, optional) = Optional Parameter. \
            All entities with a score < threshold are filtered out from the output. \
            Its best not to specify a threshold and just use the default value instead.

        Returns:
            dict : Dictionary containing keys and values of the entity type,
            ontology (such as NCBI, MeSH), ontology ID (such as the MeSH ID), the score (confidence score), and synonyms if any
        Raises:
            requestException : Invalid Request
        """
        data = {
            "mention": {
                "keyword": mention,
                "entity_type": entity_type,
                "threshold": threshold,
            }
        }

        if context:
            data["context"] = context
        output = self._perform_inference("normalizer", data)
        if output.get("errors", []):
            title, detail = self._handle_errors(output)
            raise RequestException(title, detail)

        if "term" not in output:
            return {
                "ontology": "CUI-less",
                "ontology_id": None,
                "name": None,
                "entity_type": entity_type,
            }

        return output.get("term", [])

    @Track.track_decorator
    def recognise_entity(
        self,
        text: str,
        threshold: Optional[float] = None,
        normalize_output: bool = False,
    ):
        """
        Run an NER model on the given text. The returned value is a list of entities along with span info.
        Users can simply recognise entities in a given text without any ontology standardisation
         (unlike the annotate_with_ontology function which normalises as well).

        Args:
            text (str): input text
            threshold (float, optional): Optional Parameter. \
            All entities with a score < threshold are filtered out from the output. \
            Its best not to specify a threshold and just use the default value instead.\
            normalize_output (bool): whether to normalize the keywords

        Returns:
            entities (List[dict]): List of spans containing the keyword, start/end index of the keyword and the entity type
        Raises:
            requestException : Invalid Request
        """
        # TODO: If text is too long, break it up into chunks small enough for biobert

        payload = {"text": text}
        if threshold:
            payload["threshold"] = threshold
        response = self._perform_inference("biobert", payload)
        if response.get("errors", []):
            title, detail = self._handle_errors(response)
            raise RequestException(title, detail)
        try:
            entities = response.get("entities", [])
        except KeyError as e:
            raise e

        # TODO: fetch this list from the server maybe?

        if normalize_output:
            for entity in entities:
                # don't call `normalize` for unsupported entity types
                if entity.get("entity_type") not in SUPPORTED_ENTITY_TYPES:
                    entity["name"] = None
                    continue
                norm = self.standardise_entity(
                    entity["keyword"], entity["entity_type"], text
                )
                if norm.get("ontology", []) == "CUI-less":
                    entity["name"] = None
                else:
                    entity["ontology_id"] = norm["ontology"] + ":" + norm["ontology_id"]
                    entity["name"] = norm["name"]
        return entities

    @Track.track_decorator
    def annotate_with_ontology(
        self,
        text: str,
    ) -> List[Tag]:
        """
        Tag a given piece of text. A "tag" is just an ontology term. Annotates with Polly supported ontologies.
        This function calls recognise_entity followed by normalize.
        Given a text, users can identify and tag entities in a text.
        Each entity/tag recognised in the text contains the name(word in the text identified), entity_type and the ontology_id.

        Args:
            text (str): Input text
        Returns:
            set of unique tags
        """

        entities = self.recognise_entity(text, normalize_output=True)
        res = {
            self.Tag(
                e.get("name", []), e.get("ontology_id", []), e.get("entity_type", [])
            )
            for e in entities
            if e.get("name")
        }
        return list(res)

    @Track.track_decorator
    def find_abbreviations(self, text: str) -> Dict[str, str]:
        """
        To run abbreviation detection separately.
        Internally calls a normaliser.

        Args:
            text (str): The string to detect abbreviations in.

        Returns:
        Dictionary with abbreviation as key and full form as value

        Raises:
            requestException : Invalid Request
        """
        data = {
            "mention": {"keyword": "dummykeyword", "entity_type": "gene"},
            "context": text,
        }

        response = self._perform_inference("normalizer", data)
        if "errors" in response:
            title, detail = self._handle_errors(response)
            raise RequestException(title, detail)
        try:
            output = response.get("abbreviations", [])
        except KeyError as e:
            raise e

        return output

    def assign_control_pert_labels(
        self, sample_metadata, columns_to_exclude=None
    ) -> pd.DataFrame:
        """Returns the sample metadata dataframe with 2 additional columns.
            is_control - whether the sample is a control sample
            control_prob - the probability that the sample is control
        Args:
            sample_metadata (DataFrame): Metadata table
            columns_to_exclude (Set[str]): Any columns which don't play any role in determining the label, e.g. sample id
        Returns:
            DataFrame : Input data frame with 2 additional columns
        Raises:
            requestException : Invalid Request

        """
        sample_metadata = sample_metadata.copy()

        if columns_to_exclude is None:
            columns_to_exclude = []

        cols = sample_metadata.columns.difference(columns_to_exclude)

        samples = sample_metadata[cols].to_dict("records")
        request_body = {"samples": samples}
        response = self._perform_inference("control-pertubation", request_body)
        response = {
            k: v for k, v in response.items() if k != "version" and k != "testing"
        }
        if "errors" in response:
            title, detail = self._handle_errors(response)
            raise RequestException(title, detail)
        try:
            output = pd.DataFrame(response)
        except KeyError as e:
            raise e
        sample_metadata["is_control"] = output["is_control"].values
        sample_metadata["control_prob"] = output["control_prob"].values
        return sample_metadata

    @Track.track_decorator
    def assign_clinical_labels(
        self,
        repo_name: str,
        dataset_ids: List[str],
        sample_ids: Optional[List[str]] = None,
    ) -> pd.DataFrame:
        """
        Returns a list of clinical or non clinical labels for the given datasets or samples.

        Arguments:
            repo_name (str): name of the repository for fetching datasets.
            dataset_ids (List[str]): dataset ids to be used for inference

        Keyword Arguments:
            sample_ids (List[str], optional): Optional Parameter. Sample ids if that is needed.

        Raises:
            RequestException: API response exception
            ParamException: Invalid parameters
            err

        Returns:
            dataframe which is a list of clinical tags for given ids
        """
        warnings.formatwarning = lambda msg, *args, **kwargs: f"WARNING: {msg}\n"

        try:
            self._clinical_model_param_checks(repo_name, dataset_ids, sample_ids)
            # evaluating the inference level based on if the user has provided sample_ids
            is_sample_tag = sample_ids is not None
            inference_level = "sample_id" if (is_sample_tag) else "dataset_id"

            sample_metadata, dataset_to_sample_id = self._fetch_metadata_from_cohort(
                repo_name=repo_name, dataset_ids=dataset_ids
            )

            clinical_model_predictions = []

            for dataset_id in sample_metadata:
                # Get output from model endpoint and structure output
                payload = {
                    "sample_metadata": sample_metadata[dataset_id],
                    "sample_id_column": "sample_id",
                    "dataset_id_column": "dataset_id",
                    "is_sample_tag": is_sample_tag,
                }

                output = self._perform_inference("clinical-classifier", payload)
                if "errors" in output:
                    title, detail = self._handle_errors(output)
                    raise RequestException(title, detail)

                output = output["clinical_predictions"]

                clinical_model_predictions += output

            # creating dataframe with inference_level and clinical_tags with values from the clinical_model_predictions
            clinical_tags = pd.DataFrame(
                {
                    inference_level: [
                        tag["tag_id"] for tag in clinical_model_predictions
                    ],
                    "clinical_tag": [
                        tag["clinical_tag"] for tag in clinical_model_predictions
                    ],
                }
            )

            clinical_tags = pd.merge(
                dataset_to_sample_id, clinical_tags, on=inference_level
            )

            clinical_tags = self._post_process_clinical_tags(
                clinical_tags, is_sample_tag, sample_ids
            )
        except Exception as err:
            raise err

        return clinical_tags



import os
from os.path import join as join_paths
from typing import Tuple, Optional, Union, List, TYPE_CHECKING, Callable
import json
import urllib
import tempfile

from polly.auth import Polly, UnauthorizedException
from polly.session import PollySession
from polly.s3_utils import copy_file
from polly.errors import AccessDeniedError
from polly.threading_utils import for_each_threaded
from polly_services.dataset import (
    validate_user_defined_id,
    handle_errors_or_warnings,
    S3ClientManager,
    is_supported_format,
    infer_file_type,
    extract_extension_from_s3_uri,
    extract_file_extension,
    Dataset,
    DatasetVersion,
    SupplementaryFile,
    is_gctoo,
    is_anndata,
)

if TYPE_CHECKING:
    from cmapPy.pandasGEXpress.GCToo import GCToo
    from anndata import AnnData


class Catalog:
    """
    Attributes:
        omixatlas_id: Omixatlas ID
    """

    omixatlas_id: str

    def __init__(self, omixatlas_id: str) -> None:
        """
        Initializes the internal data Catalog for a given OmixAtlas

        Args:
            omixatlas_id: The identifier for the OmixAtlas

        Examples:
            >>> catalog = Catalog(omixatlas_id='9')
        """
        if Polly.default_session is None:
            raise UnauthorizedException()

        # Instead of storing Polly.default_session directly as a private attribute
        # I'm creating a function that returns Polly.default_session
        # This ensures that any changes to default_session (e.g. token update)
        # are incorporated by an existing catalog object
        self._get_session: Callable[[], PollySession] = lambda: Polly.default_session

        self.omixatlas_id = omixatlas_id
        self._s3_client_manager = S3ClientManager(self.omixatlas_id)

    def _upload_data(
        self,
        dataset_id: str,
        data: Union[str, "GCToo", "AnnData"],
        data_format: Optional[str] = None,
    ) -> Tuple[str, str]:
        client = self._s3_client_manager.get_client()
        bucket, prefix = self._s3_client_manager.get_upload_path()

        dest_path_without_extension = "s3://" + join_paths(
            bucket, prefix, f"v2/{dataset_id}/{dataset_id}"
        )

        if isinstance(data, str) and data.startswith("s3://"):
            # user is passing s3 uri directly
            if data_format is None:
                raise ValueError(
                    "'data_format' is a required field when passing S3 uri directly"
                )
            return data, data_format

        if isinstance(data, str) and not is_supported_format(data):
            # couldn't infer format
            if data_format is None:
                raise ValueError(
                    "Couldn't infer 'data_format', please pass it directly"
                )
            s3_uri = dest_path_without_extension + extract_file_extension(data)
            s3_uri_with_version = copy_file(data, s3_uri, s3_client=client)
            return s3_uri_with_version, data_format

        if isinstance(data, str):
            # try to infer file format
            file_type, extension = infer_file_type(data)
            s3_uri = dest_path_without_extension + extension
            s3_uri_with_version = copy_file(data, s3_uri, s3_client=client)
            return s3_uri_with_version, file_type

        if is_gctoo(data):
            extension = ".gct"
            import cmapPy.pandasGEXpress.write_gct as wg

            with tempfile.NamedTemporaryFile("w", suffix=extension, delete=True) as f:
                # Using a context manager to handle the temporary file
                tmp_file_path = f.name
                wg.write(data, tmp_file_path)
                s3_uri_with_version = copy_file(
                    tmp_file_path,
                    dest_path_without_extension + extension,
                    s3_client=client,
                )
                return s3_uri_with_version, "gct"

        if is_anndata(data):
            extension = ".h5ad"
            with tempfile.NamedTemporaryFile("w", suffix=extension, delete=True) as f:
                # Using a context manager to handle the temporary file
                tmp_file_path = f.name
                data.write(tmp_file_path, compression="lzf")
                s3_uri_with_version = copy_file(
                    tmp_file_path,
                    dest_path_without_extension + extension,
                    s3_client=client,
                )
                return s3_uri_with_version, "h5ad"

        raise ValueError(f"Unsupported type ({type(data)}) for parameter 'data'")

    # TODO: support passing Path objects directly in data
    def create_dataset(
        self,
        dataset_id: str,
        data_type: str,
        data: Union[str, "GCToo", "AnnData"],
        metadata: dict,
        *,
        ingestion_params: Optional[dict] = None,
        other_params: Optional[dict] = None,
        **kwargs,
    ) -> Dataset:
        """
        Creates a new dataset in the catalog.

        Raises an error if a dataset with the same ID already exists

        Args:
            dataset_id: The unique identifier for the dataset.
            data_type: The type of data being uploaded (e.g. `Single Cell RNASeq`).
            data: The data, either as a file or as an AnnData or GCToo object
            metadata: The metadata dictionary

        Returns:
            Newly created dataset.

        Examples:
            >>> new_dataset = catalog.create_dataset(
            ...     dataset_id='GSE123',
            ...     data_type='Bulk RNAseq',
            ...     data='path/to/data/file.gct',
            ...     metadata={'description': 'New dataset description'}
            ... )
        """

        validate_user_defined_id(dataset_id, "dataset_id")

        self._check_s3_write_access()
        s3_uri_with_version, file_type = self._upload_data(
            dataset_id, data, data_format=kwargs.pop("data_format", None)
        )

        other_params = other_params or {}
        payload = {
            "data": {
                "type": "datasets",
                "id": dataset_id,
                "attributes": {
                    "data_type": data_type,
                    "metadata": metadata,
                    "data_s3_uri": s3_uri_with_version,
                    "data_format": file_type,
                    **kwargs,
                },
            },
            "meta": {"ingestion_params": ingestion_params, **other_params},
        }
        resp = self._get_session().post(
            f"{self._get_session().discover_url}/repositories/{self.omixatlas_id}/datasets",
            json=payload,
        )
        handle_errors_or_warnings(resp)
        return Dataset(**resp.json()["data"]["attributes"])

    def get_dataset(self, dataset_id: str) -> Dataset:
        """
        Retrieves the dataset with the given dataset_id if it exists

        This function doesn't download the underlying data, it only returns a reference to it

        Args:
            dataset_id: The unique identifier for the dataset to retrieve.

        Returns:
            An object representing the retrieved dataset.

        Examples:
            >>> dataset = catalog.get_dataset(dataset_id='GSE123')
            >>> dataset
            Dataset(dataset_id='GSE123', data_type='Bulk RNAseq', last_modified_at='2023-11-07 15:42:40', data_format='gct')
        """
        resp = self._get_session().get(
            f"{self._get_session().discover_url}/repositories/{self.omixatlas_id}/datasets/{dataset_id}",
        )
        handle_errors_or_warnings(resp)
        data = resp.json()["data"]
        dataset_id = data["id"]
        data["attributes"].pop("dataset_id", None)
        return Dataset(dataset_id=dataset_id, **data["attributes"])

    def list_datasets(
        self, *, limit: Optional[int] = None, prefetch_metadata: bool = False
    ) -> List[Dataset]:
        """
        Retrieves the complete list of datasets in the catalog.

        If `prefetch_metadata` is `True` the metadata json is downloaded in advance
        using multiple threads. This makes the `dataset.metadata()` function call
        return instantly. This is useful for bulk metadata download.

        Args:
            limit: Limits the number of datasets that are returned
            prefetch_metadata: Prefetches metadata for each dataset

        Returns:
            A list of Dataset instances.

        Examples:
            >>> datasets = catalog.list_datasets()
            >>> len(datasets)
            148
        """
        dataset_dicts = []

        next_link = f"/repositories/{self.omixatlas_id}/datasets"
        if limit is not None:
            # A bit hacky, `page[size]` is not the same as `limit`
            # This may not work for large values of `limit`
            next_link = f"/repositories/{self.omixatlas_id}/datasets?page[size]={limit}"

        while True:  # loop over pages
            resp = self._get_session().get(
                f"{self._get_session().discover_url}{next_link}"
            )
            handle_errors_or_warnings(resp)
            resp_dict = resp.json()
            dataset_dicts.extend(resp.json()["data"])

            if limit is not None and len(dataset_dicts) >= limit:
                dataset_dicts = dataset_dicts[:limit]
                break

            if "links" not in resp_dict or "next" not in resp_dict["links"]:
                break

            if resp_dict["links"]["next"] is None:
                break

            next_link = resp_dict["links"]["next"]

        datasets = []
        for dataset_dict in dataset_dicts:
            dataset_id = dataset_dict["id"]
            dataset_dict["attributes"].pop("dataset_id", None)
            datasets.append(
                Dataset(dataset_id=dataset_id, **dataset_dict["attributes"])
            )

        if prefetch_metadata:
            for_each_threaded(
                datasets, lambda d: d.metadata(), max_workers=20, verbose=True
            )

        return datasets

    def update_dataset(
        self,
        dataset_id: str,
        data_type: Optional[str] = None,
        metadata: Optional[dict] = None,
        data: Union[str, "GCToo", "AnnData", None] = None,
        *,
        ingestion_params: Optional[dict] = None,
        other_params: Optional[dict] = None,
        **kwargs,
    ) -> Dataset:
        """
        Updates one or more attributes of a dataset.

        Every update creates a new version of the dataset.

        Args:
            dataset_id: Identifier for the dataset.
            data_type: The type of data being uploaded (e.g. `Single Cell RNASeq`).
            data: The data, either as a file or as an AnnData or GCToo object
            metadata: The metadata dictionary

        Returns:
            The updated dataset

        Examples:
            >>> updated_dataset = catalog.update_dataset(
            ...     dataset_id='GSE123',
            ...     metadata={'new': 'metadata'}
            ... )
        """
        attrs = {**kwargs}

        if data is not None:
            self._check_s3_write_access()
            s3_uri_with_version, file_type = self._upload_data(
                dataset_id, data, data_format=kwargs.pop("data_format", None)
            )
            attrs["data_s3_uri"] = s3_uri_with_version
            attrs["data_format"] = file_type

        if data_type is not None:
            attrs["data_type"] = data_type

        if metadata is not None:
            attrs["metadata"] = metadata

        other_params = other_params or {}
        payload = {
            "data": {
                "type": "datasets",
                "id": dataset_id,
                "attributes": attrs,
            },
            "meta": {"ingestion_params": ingestion_params, **other_params},
        }

        if not attrs:
            raise ValueError("Nothing to update")

        resp = self._get_session().patch(
            f"{self._get_session().discover_url}/repositories/{self.omixatlas_id}/datasets/{dataset_id}",
            json=payload,
        )

        handle_errors_or_warnings(resp)
        data = resp.json()["data"]
        dataset_id = data["id"]
        data["attributes"].pop("dataset_id", None)
        return Dataset(dataset_id=dataset_id, **data["attributes"])

    def delete_dataset(self, dataset_id: str, **kwargs) -> None:
        """
        Performs a soft deletion of the dataset

        Args:
            dataset_id: Identifier for the dataset to be deleted.

        Examples:
            >>> catalog.delete_dataset(dataset_id='GSE123')
        """
        params = urllib.parse.quote(json.dumps(kwargs))
        resp = self._get_session().delete(
            f"{self._get_session().discover_url}/repositories/{self.omixatlas_id}/datasets/{dataset_id}?params={params}",
        )
        handle_errors_or_warnings(resp)

    def set_parent_study(
        self,
        dataset_id: str,
        study_id: Optional[str],
        *,
        ingestion_params: Optional[dict] = None,
        other_params: Optional[dict] = None,
        **kwargs,
    ) -> None:
        """
        Adds a dataset to a study. The study_id can be any arbitrary string.

        To remove the dataset from the study, set the study_id to None.

        Args:
            dataset_id: Identifier for the dataset
            study_id: User defined ID for the parent study

        Examples:
            Add dataset to a study
            >>> catalog.set_parent_study(dataset_id='GSE123', study_id='PMID123')

            Remove the dataset from the study
            >>> catalog.set_parent_study(dataset_id='GSE123', study_id=None)
        """
        if study_id is not None:
            validate_user_defined_id(study_id, "study_id")

        other_params = other_params or {}
        payload = {
            "data": {
                "type": "datasets",
                "id": dataset_id,
                "attributes": {"study_id": study_id, **kwargs},
            },
            "meta": {"ingestion_params": ingestion_params, **other_params},
        }

        resp = self._get_session().patch(
            f"{self._get_session().discover_url}/repositories/{self.omixatlas_id}/datasets/{dataset_id}",
            json=payload,
        )

        handle_errors_or_warnings(resp)

    def list_datasets_in_study(self, study_id: str) -> List[str]:
        """
        Retrieves a list of datasets associated with a specific study.

        Args:
            study_id: Identifier for the study.

        Returns:
            A list of dataset IDs that are part of the study.

        Examples:
            >>> catalog.list_datasets_in_study(study_id='PMID123')
            ['GSE123', 'GSE123_raw']

        """
        resp = self._get_session().get(
            f"{self._get_session().discover_url}/repositories/{self.omixatlas_id}/studies/{study_id}",
        )

        handle_errors_or_warnings(resp)
        items = resp.json()["relationships"]["datasets"]["data"]
        return [item["id"] for item in items]

    def list_studies(self) -> List[str]:
        """
        Retrieves a list of all studies in the catalog.

        Returns:
            A list of study IDs.

        Examples:
            >>> study_ids = catalog.list_studies()
            >>> study_ids
            ['PMID123', 'PMID125', 'SCP300']
        """
        resp = self._get_session().get(
            f"{self._get_session().discover_url}/repositories/{self.omixatlas_id}/studies"
        )
        handle_errors_or_warnings(resp)
        studies = resp.json()["data"]
        return [study["id"] for study in studies]

    def get_version(self, dataset_id: str, version_id: str) -> DatasetVersion:
        """
        Retrieves a specific version of a dataset.

        You can also use this to retrieve a version that may have been deleted.

        Args:
            dataset_id (str): Identifier for the dataset.
            version_id (str): Identifier for the dataset version.

        Returns:
            An object representing the dataset version.

        Examples:
            >>> catalog.get_version(dataset_id='GSE123', version_id='<uuid>')
            DatasetVersion(version_id='<uuid>', created_at='2023-10-29 20:41:15')
        """
        resp = self._get_session().get(
            f"{self._get_session().discover_url}/repositories/{self.omixatlas_id}/datasets/{dataset_id}/versions/{version_id}",
        )
        handle_errors_or_warnings(resp)
        data = resp.json()["data"]
        version_id = data["id"]
        data["attributes"].pop("version_id", None)
        return DatasetVersion(version_id=version_id, **data["attributes"])

    def list_versions(
        self, dataset_id: str, *, include_deleted: bool = False
    ) -> List[DatasetVersion]:
        """
        Lists dataset versions, optionally including deleted ones.

        Args:
            dataset_id: Identifier for the dataset.
            include_deleted: If set to True, includes previously deleted versions

        Returns:
            List of dataset versions.
        """
        resp = self._get_session().get(
            f"{self._get_session().discover_url}/repositories/{self.omixatlas_id}/"
            f"datasets/{dataset_id}/versions?include_deleted={include_deleted}"
        )
        handle_errors_or_warnings(resp)
        data = resp.json()["data"]
        versions = []
        for version_dict in data:
            version_id = version_dict["id"]
            version_dict["attributes"].pop("version_id", None)
            versions.append(
                DatasetVersion(version_id=version_id, **version_dict["attributes"])
            )
        return versions

    def rollback(
        self,
        dataset_id: str,
        version_id: str,
        *,
        ingestion_params: Optional[dict] = None,
        other_params: Optional[dict] = None,
        **kwargs,
    ) -> Dataset:
        """
        Reverts a dataset to a specified older version.

        This function updates the dataset's `data` and `metadata` to match the specified
        version. The version ID itself is not restored.

        Args:
            dataset_id: Identifier for the dataset.
            version_id: Identifier for the version to roll back to.

        Returns:
            The dataset instance with the rolled back data and metadata.
        """
        version = self.get_version(dataset_id, version_id)

        if dataset_id in self:
            return self.update_dataset(
                dataset_id,
                data_type=version.data_type,
                metadata=version.metadata(),
                data=version.data_location,
                data_format=version.data_format,  # important: don't try to infer format
                ingestion_params=ingestion_params,
                other_params=other_params,
                **kwargs,
            )
        else:
            return self.create_dataset(
                dataset_id,
                data_type=version.data_type,
                data=version.data_location,
                metadata=version.metadata(),
                data_format=version.data_format,  # important: don't try to infer format
                ingestion_params=ingestion_params,
                other_params=other_params,
                **kwargs,
            )

    def attach_file(
        self,
        dataset_id: str,
        path: str,
        *,
        tag: Optional[str] = None,
        source_version_id: Optional[str] = None,
        file_name: Optional[str] = None,
        **kwargs: dict,
    ) -> SupplementaryFile:
        """
        Attach a supplementary file to a dataset.

        If a file with the same name already exists it is overwritten.

        Optionally, you can link the supplementary file to a specific version of a dataset.
        This will auto-delete the file if the underlying data for the dataset changes.
        Note that changes to `metadata` or `data_type` will not auto-delete the file.

        _Added in polly-python v1.3_

        Args:
            dataset_id: Identifier for the dataset.
            path: Local path of the file.
            tag: An optional tag for the supplementary file.
            source_version_id: The version_id of the source dataset.
            file_name: The name of the file. If not provided the file name will be
              inferred from the path.

        Returns:
            A SupplementaryFile instance representing the attached file.
        """
        self._check_s3_write_access()

        # user is passing s3 uri directly
        if path.startswith("s3://"):
            err_msg = (
                "The 'file_format' and 'file_name' need to be"
                " explicitly provided if path is an s3 uri"
            )

            if "file_format" not in kwargs:
                raise ValueError(err_msg)
            if file_name is None:
                raise ValueError(err_msg)

            file_format = kwargs["file_format"]
            s3_uri_with_version = path

        else:
            file_name = file_name or os.path.basename(path)
            file_format = kwargs.get("file_format")

            if file_format is None and is_supported_format(path):
                file_format, _ = infer_file_type(path)
            elif file_format is None:
                raise ValueError(
                    "Could not infer the file format, please provide it"
                    " explicitly using the 'file_format' parameter"
                )

            client = self._s3_client_manager.get_client()
            bucket, prefix = self._s3_client_manager.get_upload_path()
            dest_path = "s3://" + join_paths(
                bucket, prefix, f"v2/{dataset_id}/.sfiles/{file_name}"
            )
            s3_uri_with_version = copy_file(path, dest_path, s3_client=client)

        payload = {
            "data": {
                "type": "supplementary_files",
                "attributes": {
                    "file_name": file_name,
                    "file_s3_uri": s3_uri_with_version,
                    "file_format": file_format,
                    "tag": tag,
                    "source_version_id": source_version_id,
                    **kwargs,
                },
            }
        }

        resp = self._get_session().post(
            f"{self._get_session().discover_url}/repositories/{self.omixatlas_id}/datasets/{dataset_id}/supplementary_files",
            json=payload,
        )

        handle_errors_or_warnings(resp)

        data = resp.json()["data"]
        return SupplementaryFile(**data["attributes"])

    def get_file(self, dataset_id: str, file_name: str) -> SupplementaryFile:
        """
        Retrieves a supplementary file.

        _Added in polly-python v1.3_

        Args:
            dataset_id: Identifier for the dataset.
            file_name: Name of the file to retrieve.

        Returns:
            A SupplementaryFile instance.
        """
        sess = self._get_session()
        resp = sess.get(
            f"{sess.discover_url}/repositories/{self.omixatlas_id}/datasets/{dataset_id}"
            f"/supplementary_files/{file_name}"
        )
        handle_errors_or_warnings(resp)
        data = resp.json()["data"]
        return SupplementaryFile(**data["attributes"])

    def list_files(self, dataset_id: str) -> List[SupplementaryFile]:
        """
        Retrieves a list of all supplementary files attached to a dataset.

        _Added in polly-python v1.3_

        Args:
            dataset_id: Identifier for the dataset.

        Returns:
            A list of SupplementaryFile instances.
        """
        sess = self._get_session()
        resp = sess.get(
            f"{sess.discover_url}/repositories/{self.omixatlas_id}/datasets/{dataset_id}"
            f"/supplementary_files"
        )
        handle_errors_or_warnings(resp)
        data = resp.json()["data"]
        return [SupplementaryFile(**item["attributes"]) for item in data]

    def delete_file(self, dataset_id: str, file_name: str) -> None:
        """
        Deletes the supplementary file.

        _Added in polly-python v1.3_

        Args:
            dataset_id: Identifier for the dataset.
            file_name: File name
        """
        sess = self._get_session()
        resp = sess.delete(
            f"{sess.discover_url}/repositories/{self.omixatlas_id}/datasets/{dataset_id}"
            f"/supplementary_files/{file_name}"
        )
        handle_errors_or_warnings(resp)

    def _check_s3_write_access(self):
        access = self._s3_client_manager.get_access_level()

        if access is None:
            return

        if access == "read":
            raise AccessDeniedError(
                detail=(
                    "You don't have permission to write to"
                    f" omixatlas_id='{self.omixatlas_id}'. "
                    "\n\nRun `catalog.clear_token_cache()` before retrying"
                ),
            )

    def clear_token_cache(self) -> None:
        """
        Clears cached S3 tokens
        """
        self._s3_client_manager.clear()

    def trigger_ingestion(
        self,
        dataset_id: str,
        *,
        ingestion_params: Optional[dict] = None,
        other_params: Optional[dict] = None,
    ):
        """
        A helper function to manually trigger ingestion for a dataset

        Args:
            dataset_id: Identifier for the dataset.
        """
        other_params = other_params or {}
        payload = {
            "data": {
                "type": "datasets",
                "id": dataset_id,
                "attributes": {},
            },
            "meta": {"ingestion_params": ingestion_params, **other_params},
        }

        resp = self._get_session().patch(
            f"{self._get_session().discover_url}/repositories/{self.omixatlas_id}/datasets/{dataset_id}",
            json=payload,
        )
        handle_errors_or_warnings(resp)
        return resp.json()

    def __contains__(self, dataset_id: str) -> bool:
        """
        Check if a dataset with the given dataset_id exists in the catalog.

        Args:
            dataset_id: Identifier for the dataset.

        Returns:
            True if the dataset exists, False otherwise.
        """
        resp = self._get_session().get(
            f"{self._get_session().discover_url}/repositories/{self.omixatlas_id}/datasets/{dataset_id}"
        )
        if resp.status_code == 404:
            return False
        handle_errors_or_warnings(resp)
        return True

    def __repr__(self):
        return f"Catalog(omixatlas_id='{self.omixatlas_id}')"


def copy_dataset(
    src_omixatlas_id: str,
    src_dataset_id: str,
    dest_omixatlas_id: str,
    dest_dataset_id: str,
    overwrite: bool = False,
    *,
    ingestion_params: Optional[dict] = None,
    other_params: Optional[dict] = None,
    **kwargs,
) -> Dataset:
    """
    Copies a dataset from one omixatlas catalog to another.

    Any supplementary files attached to the dataset are also copied.

    The transfer happens remotely, the data is not downloaded locally.

    Args:
        src_omixatlas_id: The source omixatlas_id
        src_dataset_id: The source dataset_id
        dest_omixatlas_id: The destination omixatlas_id
        dest_dataset_id: The destination dataset_id
        overwrite: If True, overwrites the destination dataset if it already exists

    Returns:
        The dataset in the destination catalog.

    Examples:
        >>> from polly.data_management import copy_dataset
        >>> copy_dataset('9', 'GSE123', '17', 'GSE123_copy')
    """
    validate_user_defined_id(dest_dataset_id, "dest_dataset_id")

    print("Copying dataset")
    src_catalog = Catalog(src_omixatlas_id)
    src_dataset = src_catalog.get_dataset(src_dataset_id)

    if src_omixatlas_id == dest_omixatlas_id:
        # when copying within the same catalog
        # the destination dataset points to the same file as source dataset
        dest_dataset = _copy_dataset(
            src_dataset,
            Catalog(dest_omixatlas_id),
            dest_dataset_id,
            dest_location=src_dataset.data_location,  # same as src location
            overwrite=overwrite,
            ingestion_params=ingestion_params,
            other_params=other_params,
            **kwargs,
        )
        print("Copying supplementary files")
        for file in src_catalog.list_files(src_dataset.dataset_id):
            _copy_supplementary_file(
                src_omixatlas_id, src_dataset, dest_omixatlas_id, dest_dataset, file
            )
        return dest_dataset

    client_manager = S3ClientManager(src_omixatlas_id, dest_omixatlas_id)
    bucket, prefix = client_manager.get_upload_path()

    dest_path_without_extension = "s3://" + join_paths(
        bucket, prefix, f"v2/{dest_dataset_id}/{dest_dataset_id}"
    )
    extension = extract_extension_from_s3_uri(src_dataset.data_location)

    s3_uri_with_version = copy_file(
        src_dataset.data_location,
        dest_path_without_extension + extension,
        s3_client=client_manager.get_client(),
    )
    dest_catalog = Catalog(dest_omixatlas_id)

    dest_dataset = _copy_dataset(
        src_dataset,
        dest_catalog,
        dest_dataset_id,
        s3_uri_with_version,
        overwrite,
        ingestion_params=ingestion_params,
        other_params=other_params,
        **kwargs,
    )
    print("Copying supplementary files")
    for file in src_catalog.list_files(src_dataset.dataset_id):
        _copy_supplementary_file(
            src_omixatlas_id, src_dataset, dest_omixatlas_id, dest_dataset, file
        )
    return dest_dataset


def _copy_dataset(
    src_dataset: Dataset,
    dest_catalog: Catalog,
    dest_dataset_id: str,
    dest_location: str,
    overwrite: bool,
    *,
    ingestion_params: Optional[dict] = None,
    other_params: Optional[dict] = None,
    **kwargs,
) -> Dataset:
    if overwrite and dest_dataset_id in dest_catalog:
        return dest_catalog.update_dataset(
            dest_dataset_id,
            src_dataset.data_type,
            metadata=src_dataset.metadata(),
            data=dest_location,
            data_format=src_dataset.data_format,
            study_id=src_dataset.study_id,
            ingestion_params=ingestion_params,
            other_params=other_params,
            **kwargs,
        )

    return dest_catalog.create_dataset(
        dest_dataset_id,
        src_dataset.data_type,
        data=dest_location,
        metadata=src_dataset.metadata(),
        data_format=src_dataset.data_format,
        study_id=src_dataset.study_id,
        ingestion_params=ingestion_params,
        other_params=other_params,
        **kwargs,
    )


def _copy_supplementary_file(
    src_catalog_id: str,
    src_dataset: Dataset,
    dest_catalog_id: Catalog,
    dest_dataset: Dataset,
    file: SupplementaryFile,
) -> SupplementaryFile:
    if src_catalog_id == dest_catalog_id:
        # pass s3 uri directly
        return Catalog(dest_catalog_id).attach_file(
            dest_dataset.dataset_id,
            path=file.file_location,
            tag=file.tag,
            file_name=file.file_name,
            file_format=file.file_format,
            source_version_id=(
                dest_dataset.version_id if file.derived_from is not None else None
            ),
        )

    client_manager = S3ClientManager(src_catalog_id, dest_catalog_id)
    bucket, prefix = client_manager.get_upload_path()

    dest_s3_uri = "s3://" + join_paths(
        bucket, prefix, f"v2/{dest_dataset.dataset_id}/.sfiles/{file.file_name}"
    )

    # Copy the file to the destination S3 URI
    copy_file(file.file_location, dest_s3_uri, s3_client=client_manager.get_client())

    # Attach the copied file to the destination dataset
    return Catalog(dest_catalog_id).attach_file(
        dest_dataset.dataset_id,
        path=dest_s3_uri,
        file_name=file.file_name,
        file_format=file.file_format,
        tag=file.tag,
        source_version_id=(
            dest_dataset.version_id if file.derived_from is not None else None
        ),
    )



"""
slow gloms that came up organically, used as performance metrics
"""
import time
import gc

import attr

from glom import glom, T



STR_SPEC = [{
    'id': ('id', str),
    'name': 'short_name',
    'external_id': 'external_id',
    'created_date': 'created_date',
}]


T_SPEC = [{
    'id': (T.id, str),
    'name': T.short_name,
    'external_id': T.external_id,
    'created_date': T.created_date,
}]


def func(data):
    return [{
            'id': str(t.id),
            'name': t.short_name,
            'external_id': t.external_id,
            'created_date': t.created_date
        } for t in data]


def setup_list_of_dict(num=100):
    """
    a common use case is list-of-dicts object processing
    to prepare internal objects for JSON serialization
    """
    Obj = attr.make_class(
        'Obj', ['id', 'short_name', 'external_id', 'created_date'])

    data = [
        Obj(i, 'name' + str(i), 'external' + str(i), 'now') for i in range(num)]

    return data


def run(spec, data):
    start = time.time()
    glom(data, spec)
    end = time.time()
    print("{} us per object".format((end - start) / len(data) * 1e6))


def ratio(spec, func, data):
    glom_dur = []
    py_dur = []
    for i in range(10):
        t1 = time.perf_counter_ns()
        glom(data, spec)
        t2 = time.perf_counter_ns()
        func(data)
        t3 = time.perf_counter_ns()
        glom_dur.append(t2 - t1)
        py_dur.append(t3 - t2)

    glom_avg = sum(sorted(glom_dur)[2:-2])
    py_avg = sum(sorted(py_dur)[2:-2])

    return 1.0 * glom_avg / py_avg


if __name__ == "__main__":
    import cProfile
    data = setup_list_of_dict(100000)
    run(STR_SPEC, data)
    run(STR_SPEC, data)
    print(ratio(STR_SPEC, func, setup_list_of_dict(1000)))
    print(ratio(STR_SPEC, func, setup_list_of_dict(1000)))


# suggest using scalene to profile with:
# $ scalene glom/test/perf_report.py --profile-all --reduced-profile --cpu-only --outfile SCALENE-CPU.txt



from typing import Dict, Type, Union

from pydantic import BaseModel


class VpcLatticeModel(BaseModel):
    method: str
    raw_path: str
    body: Union[str, Type[BaseModel]]
    is_base64_encoded: bool
    headers: Dict[str, str]
    query_string_parameters: Dict[str, str]



from dataclasses import asdict, is_dataclass
from typing import Any, Dict, Type

from aws_lambda_powertools.utilities.idempotency.exceptions import (
    IdempotencyModelTypeError,
    IdempotencyNoSerializationModelError,
)
from aws_lambda_powertools.utilities.idempotency.serialization.base import (
    BaseIdempotencyModelSerializer,
    BaseIdempotencySerializer,
)

DataClass = Any


class DataclassSerializer(BaseIdempotencyModelSerializer):
    """
    A serializer class for transforming data between dataclass objects and dictionaries.
    """

    def __init__(self, model: Type[DataClass]):
        """
        Parameters
        ----------
        model: Type[DataClass]
            A dataclass type to be used for serialization and deserialization
        """
        self.__model: Type[DataClass] = model

    def to_dict(self, data: DataClass) -> Dict:
        return asdict(data)

    def from_dict(self, data: Dict) -> DataClass:
        return self.__model(**data)

    @classmethod
    def instantiate(cls, model_type: Any) -> BaseIdempotencySerializer:
        if model_type is None:
            raise IdempotencyNoSerializationModelError("No serialization model was supplied")

        if not is_dataclass(model_type):
            raise IdempotencyModelTypeError("Model type is not inherited of dataclass type")
        return cls(model=model_type)



from glom import Auto, Fill, T, glom

def test():
    assert glom('abc', Fill((T[0], {T[1]: T[2]}))) == ('a', {'b': 'c'})
    assert glom('123', Fill({T[0], frozenset([T[1], T[2]])})) == {'1', frozenset(['2', '3'])}
    assert glom('xyz', Fill([T[0], T[1], T[2]]))
    assert glom('abc', Fill(lambda t: t.upper())) == 'ABC'
    assert glom('a', Fill(1)) == 1
    assert Fill((T, T, T)).fill(1) == (1, 1, 1)

    target = {'data': [0, 2, 4]}
    assert glom(target, Fill((T['data'][-1], Auto('data.-2')))) == (4, 2)

    assert repr(Auto()) == 'Auto()'
    assert repr(Auto(T)) == 'Auto(T)'

    assert repr(Fill()) == 'Fill()'
    assert repr(Fill(T)) == 'Fill(T)'

    assert repr(Fill(len)) == 'Fill(len)'



from aws_lambda_powertools.metrics.provider.datadog.datadog import DatadogProvider
from aws_lambda_powertools.metrics.provider.datadog.metrics import DatadogMetrics

__all__ = [
    "DatadogMetrics",
    "DatadogProvider",
]



import ast
import builtins
import compileall
import importlib
import os
import runpy
import shutil
import subprocess
import sys
import tarfile
import typing as t
from contextlib import contextmanager, suppress
from functools import partial
from pathlib import Path
from types import FrameType
from typing import List, Optional

import datapane.client.api as api
from datapane.common import SDict, SSDict, log

from .exceptions import CodeError, CodeRaisedError, CodeSyntaxError

if t.TYPE_CHECKING:
    # for BC reasons
    LegacyApp = t.Any


ENVIRON_CONFIG = {
    "banned_builtins": {"compile", "exec", "eval"},
    "default_environment": [],
}
RUN_NAME = "__datapane__"  # <datapane> ??


def run(app: LegacyApp, user_config: SDict, env: SSDict) -> SDict:
    """Run a datapane python script/module"""
    api._reset_runtime(params=user_config)
    # use the app id for unique, isolated dir per app
    env_dir = Path(app.id)
    setup_script(app, env_dir)

    with script_env(env_dir, env):
        # app_name = str(app.script) ## <module> ?
        try:
            # run script in unpacked dir wrapped around pre/post-commands
            run_commands(app.pre_commands)
            res_scope = exec_mod(app.script)
            run_commands(app.post_commands)
            return res_scope
        except SyntaxError:
            raise CodeSyntaxError.from_exception()
        except Exception:
            raise CodeRaisedError.from_exception(partial(filter_frame_by_filename, "<module>"))


# OBSOLETE
# def run(run_config: RunnerConfig) -> List[api.ReportBlock]:
#     """Snippet - run a python function embedded within in the snippet config field"""
#     code = run_config.code
#     user_config: SDict = Munch.fromDict(run_config.format())
#     # call to_df locally
#     # convert to dfs if needed
#     # dp._snippet_init(config=user_config)
#
#     # NOTE(MG): currently just exec the script - do we need to load it as a module via importlib?
#     try:
#         # pass in 'snippet_name' explicitly as we depend on it below.
#         init_state: SDict = {"params": user_config, "parameters": user_config, "on_datapane": True}
#         res_scope = exec_user_script(code, init_state, snippet_name=USER_CODE_NAME)
#
#         # call render function/extract report var
#         if "render" in res_scope:
#             report = res_scope["render"]()
#         elif "report" in res_scope:
#             report = res_scope["report"]
#         else:
#             log.warning("report variable nor render function found in script")
#             report = []
#         return report
#     except SyntaxError:
#         raise CodeSyntaxError.from_exception()
#     except Exception:
#         raise CodeRaisedError.from_exception(partial(filter_frame_by_filename, USER_CODE_NAME))


################################################################################
# internal
def in_venv() -> bool:
    return hasattr(sys, "real_prefix") or sys.base_prefix != sys.prefix


@contextmanager
def script_env(env_dir: Path, env: SSDict) -> t.Generator[None, None, None]:
    """
    Change the local dir and add to site-path so relative files and imports work
    TODO
        - this is NOT thread-safe - unlikely we can run multiple concurrent scripts atm
        - this doesn't save env's as a call-stack directly - however handled implictly via Python stack anyway
    """
    cwd = os.getcwd()
    log.debug(f"[cd] {cwd} -> {env_dir}")
    if not env_dir.exists():
        env_dir.mkdir(parents=True)

    full_env_dir = str(env_dir.resolve())
    sys.path.insert(0, full_env_dir)
    os.chdir(full_env_dir)
    os.environ.update(env)
    try:
        yield
    finally:
        for env_key in env.keys():
            with suppress(KeyError):
                os.environ.pop(env_key)
        try:
            sys.path.remove(full_env_dir)
        except ValueError as e:
            raise CodeError("sys.path not as expected - was it modified?") from e
        os.chdir(cwd)

        log.debug(f"[cd] {cwd} <- {env_dir}")
        # shutil.rmtree(env_dir, ignore_errors=True)


def run_commands(cmds: List[str]) -> None:
    if cmds:
        _cmds = "\n".join(cmds)
        subprocess.run(args=_cmds, check=True, shell=True)
        if "pip" in _cmds:
            importlib.invalidate_caches()  # ensure new packages are detected


def setup_script(s: LegacyApp, env_dir: Path):
    """Setup the script - unpack & install deps"""
    # TODO - add local cache check here
    if env_dir.exists():
        log.debug("Package already exists, not redownloading")
        return None

    # download and unpack bundle locally into env_dir
    sdist = s.download_pkg()
    assert tarfile.is_tarfile(sdist), "Invalid sdist file"
    shutil.unpack_archive(sdist, extract_dir=env_dir, format="gztar")
    sdist.unlink()
    comp_r = compileall.compile_dir(env_dir, force=True, workers=1, quiet=1)
    if not comp_r:
        log.warning("Compiling app bundle failed - errors may occur")

    # install deps
    if s.requirements:
        pip_args = [sys.executable, "-m", "pip", "install"]
        if os.getuid() != 0 and not in_venv():  # type: ignore [attr-defined]
            # we're a normal/non-root user outside a venv
            pip_args.append("--user")
        pip_args.extend(s.requirements)
        log.debug(f"Calling pip as '{pip_args}'")
        subprocess.run(args=pip_args, check=True)
        importlib.invalidate_caches()  # ensure new packages are detected

    log.info(f"Successfully installed bundle for app {s.id}")


def exec_mod(script_path: Path, init_state: Optional[SDict] = None) -> SDict:
    # a = ast.parse(app, snippet_name, "exec")
    # ast_validation(a)
    init_state = init_state or dict()
    globalscope = {
        # "df": single_input,
        # "params": Munch(config),
        # "__builtins__": override_builtins(__import__)
    }
    globalscope.update(init_state)

    # we're using run_py rather than run_module to ensure same env as running on user's machine locally
    # we have a script, not a module - must follow script semantics
    res_scope = runpy.run_path(str(script_path), init_globals=globalscope, run_name=RUN_NAME)
    return res_scope


def ast_validation(node: ast.AST):
    for n in ast.walk(node):
        # TODO: implement check for import statements
        if isinstance(n, ast.Import):
            pass
        if isinstance(n, ast.ImportFrom):
            pass


class OverriddenBuiltins(dict):
    def __getitem__(self, name):  # noqa: ANN001
        if name in ENVIRON_CONFIG["banned_builtins"]:
            raise RuntimeError(f"Illegal builtin {name}")
        return super().__getitem__(name)


def override_builtins(importer) -> OverriddenBuiltins:  # noqa: ANN001
    b = OverriddenBuiltins(builtins.__dict__)
    b["__import__"] = importer
    return b


def importer(name, *x, **y):  # noqa: ANN001, ANN201
    rootpkg = name.split(".")[0]
    if rootpkg not in ("runner",):  # TODO: module check
        return builtins.__import__(name, *x, **y)
    raise ImportError("Cannot import banned module")


def filter_frame_by_filename(filename: str, frame: FrameType) -> bool:
    return frame.f_code.co_filename == filename



import json
from typing import Optional, Union

from aws_lambda_powertools.utilities._data_masking.provider import BaseProvider


class DataMasking:
    """
    Note: This utility is currently in a Non-General Availability (Non-GA) phase and may have limitations.
    Please DON'T USE THIS utility in production environments.
    Keep in mind that when we transition to General Availability (GA), there might be breaking changes introduced.

    A utility class for masking sensitive data within various data types.

    This class provides methods for masking sensitive information, such as personal
    identifiers or confidential data, within different data types such as strings,
    dictionaries, lists, and more. It helps protect sensitive information while
    preserving the structure of the original data.

    Usage:
    Instantiate an object of this class and use its methods to mask sensitive data
    based on the data type. Supported data types include strings, dictionaries,
    and more.

    Example:
    ```
    from aws_lambda_powertools.utilities.data_masking.base import DataMasking

    def lambda_handler(event, context):
        masker = DataMasking()

        data = {
            "project": "powertools",
            "sensitive": "xxxxxxxxxx"
        }

        masked = masker.mask(data,fields=["sensitive"])

        return masked

    ```
    """

    def __init__(self, provider: Optional[BaseProvider] = None):
        self.provider = provider or BaseProvider()

    def encrypt(self, data, fields=None, **provider_options):
        return self._apply_action(data, fields, self.provider.encrypt, **provider_options)

    def decrypt(self, data, fields=None, **provider_options):
        return self._apply_action(data, fields, self.provider.decrypt, **provider_options)

    def mask(self, data, fields=None, **provider_options):
        return self._apply_action(data, fields, self.provider.mask, **provider_options)

    def _apply_action(self, data, fields, action, **provider_options):
        """
        Helper method to determine whether to apply a given action to the entire input data
        or to specific fields if the 'fields' argument is specified.

        Parameters
        ----------
        data : any
            The input data to process.
        fields : Optional[List[any]] = None
            A list of fields to apply the action to. If 'None', the action is applied to the entire 'data'.
        action : Callable
           The action to apply to the data. It should be a callable that performs an operation on the data
           and returns the modified value.

        Returns
        -------
        any
            The modified data after applying the action.
        """

        if fields is not None:
            return self._apply_action_to_fields(data, fields, action, **provider_options)
        else:
            return action(data, **provider_options)

    def _apply_action_to_fields(
        self,
        data: Union[dict, str],
        fields: list,
        action,
        **provider_options,
    ) -> Union[dict, str]:
        """
        This method takes the input data, which can be either a dictionary or a JSON string,
        and applies a mask, an encryption, or a decryption to the specified fields.

        Parameters
        ----------
            data : Union[dict, str])
                The input data to process. It can be either a dictionary or a JSON string.
            fields : List
                A list of fields to apply the action to. Each field can be specified as a string or
                a list of strings representing nested keys in the dictionary.
            action : Callable
                The action to apply to the fields. It should be a callable that takes the current
                value of the field as the first argument and any additional arguments that might be required
                for the action. It performs an operation on the current value using the provided arguments and
                returns the modified value.
            **provider_options:
                Additional keyword arguments to pass to the 'action' function.

        Returns
        -------
            dict
                The modified dictionary after applying the action to the
            specified fields.

        Raises
        -------
            ValueError
                If 'fields' parameter is None.
            TypeError
                If the 'data' parameter is not a traversable type

        Example
        -------
        ```python
        >>> data = {'a': {'b': {'c': 1}}, 'x': {'y': 2}}
        >>> fields = ['a.b.c', 'a.x.y']
        # The function will transform the value at 'a.b.c' (1) and 'a.x.y' (2)
        # and store the result as:
        new_dict = {'a': {'b': {'c': 'transformed_value'}}, 'x': {'y': 'transformed_value'}}
        ```
        """

        if fields is None:
            raise ValueError("No fields specified.")

        if isinstance(data, str):
            # Parse JSON string as dictionary
            my_dict_parsed = json.loads(data)
        elif isinstance(data, dict):
            # In case their data has keys that are not strings (i.e. ints), convert it all into a JSON string
            my_dict_parsed = json.dumps(data)
            # Turn back into dict so can parse it
            my_dict_parsed = json.loads(my_dict_parsed)
        else:
            raise TypeError(
                f"Unsupported data type for 'data' parameter. Expected a traversable type, but got {type(data)}.",
            )

        # For example: ['a.b.c'] in ['a.b.c', 'a.x.y']
        for nested_key in fields:
            # Prevent overriding loop variable
            curr_nested_key = nested_key

            # If the nested_key is not a string, convert it to a string representation
            if not isinstance(curr_nested_key, str):
                curr_nested_key = json.dumps(curr_nested_key)

            # Split the nested key string into a list of nested keys
            # ['a.b.c'] -> ['a', 'b', 'c']
            keys = curr_nested_key.split(".")

            # Initialize a current dictionary to the root dictionary
            curr_dict = my_dict_parsed

            # Traverse the dictionary hierarchy by iterating through the list of nested keys
            for key in keys[:-1]:
                curr_dict = curr_dict[key]

            # Retrieve the final value of the nested field
            valtochange = curr_dict[(keys[-1])]

            # Apply the specified 'action' to the target value
            curr_dict[keys[-1]] = action(valtochange, **provider_options)

        return my_dict_parsed



import dataclasses
import datetime
from collections import defaultdict, deque
from decimal import Decimal
from enum import Enum
from pathlib import Path, PurePath
from re import Pattern
from types import GeneratorType
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
from uuid import UUID

from pydantic import BaseModel
from pydantic.color import Color
from pydantic.types import SecretBytes, SecretStr

from aws_lambda_powertools.event_handler.openapi.compat import _model_dump
from aws_lambda_powertools.event_handler.openapi.types import IncEx

"""
This module contains the encoders used by jsonable_encoder to convert Python objects to JSON serializable data types.
"""


def jsonable_encoder(  # noqa: PLR0911
    obj: Any,
    include: Optional[IncEx] = None,
    exclude: Optional[IncEx] = None,
    by_alias: bool = True,
    exclude_unset: bool = False,
    exclude_defaults: bool = False,
    exclude_none: bool = False,
) -> Any:
    """
    JSON encodes an arbitrary Python object into JSON serializable data types.

    This is a modified version of fastapi.encoders.jsonable_encoder that supports
    encoding of pydantic.BaseModel objects.

    Parameters
    ----------
    obj : Any
        The object to encode
    include : Optional[IncEx], optional
        A set or dictionary of strings that specifies which properties should be included, by default None,
        meaning everything is included
    exclude : Optional[IncEx], optional
        A set or dictionary of strings that specifies which properties should be excluded, by default None,
        meaning nothing is excluded
    by_alias : bool, optional
        Whether field aliases should be respected, by default True
    exclude_unset : bool, optional
        Whether fields that are not set should be excluded, by default False
    exclude_defaults : bool, optional
        Whether fields that are equal to their default value (as specified in the model) should be excluded,
        by default False
    exclude_none : bool, optional
        Whether fields that are equal to None should be excluded, by default False

    Returns
    -------
    Any
        The JSON serializable data types
    """
    if include is not None and not isinstance(include, (set, dict)):
        include = set(include)
    if exclude is not None and not isinstance(exclude, (set, dict)):
        exclude = set(exclude)

    # Pydantic models
    if isinstance(obj, BaseModel):
        return _dump_base_model(
            obj=obj,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
        )

    # Dataclasses
    if dataclasses.is_dataclass(obj):
        obj_dict = dataclasses.asdict(obj)
        return jsonable_encoder(
            obj_dict,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

    # Enums
    if isinstance(obj, Enum):
        return obj.value

    # Paths
    if isinstance(obj, PurePath):
        return str(obj)

    # Scalars
    if isinstance(obj, (str, int, float, type(None))):
        return obj

    # Dictionaries
    if isinstance(obj, dict):
        return _dump_dict(
            obj=obj,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_none=exclude_none,
            exclude_unset=exclude_unset,
        )

    # Sequences
    if isinstance(obj, (list, set, frozenset, GeneratorType, tuple, deque)):
        return _dump_sequence(
            obj=obj,
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_none=exclude_none,
            exclude_defaults=exclude_defaults,
            exclude_unset=exclude_unset,
        )

    # Other types
    if type(obj) in ENCODERS_BY_TYPE:
        return ENCODERS_BY_TYPE[type(obj)](obj)

    for encoder, classes_tuple in encoders_by_class_tuples.items():
        if isinstance(obj, classes_tuple):
            return encoder(obj)

    # Default
    return _dump_other(
        obj=obj,
        include=include,
        exclude=exclude,
        by_alias=by_alias,
        exclude_none=exclude_none,
        exclude_unset=exclude_unset,
        exclude_defaults=exclude_defaults,
    )


def _dump_base_model(
    *,
    obj: Any,
    include: Optional[IncEx] = None,
    exclude: Optional[IncEx] = None,
    by_alias: bool = True,
    exclude_unset: bool = False,
    exclude_none: bool = False,
    exclude_defaults: bool = False,
):
    """
    Dump a BaseModel object to a dict, using the same parameters as jsonable_encoder
    """
    obj_dict = _model_dump(
        obj,
        mode="json",
        include=include,
        exclude=exclude,
        by_alias=by_alias,
        exclude_unset=exclude_unset,
        exclude_none=exclude_none,
        exclude_defaults=exclude_defaults,
    )
    if "__root__" in obj_dict:
        obj_dict = obj_dict["__root__"]

    return jsonable_encoder(
        obj_dict,
        exclude_none=exclude_none,
        exclude_defaults=exclude_defaults,
    )


def _dump_dict(
    *,
    obj: Any,
    include: Optional[IncEx] = None,
    exclude: Optional[IncEx] = None,
    by_alias: bool = True,
    exclude_unset: bool = False,
    exclude_none: bool = False,
) -> Dict[str, Any]:
    """
    Dump a dict to a dict, using the same parameters as jsonable_encoder
    """
    encoded_dict = {}
    allowed_keys = set(obj.keys())
    if include is not None:
        allowed_keys &= set(include)
    if exclude is not None:
        allowed_keys -= set(exclude)
    for key, value in obj.items():
        if (
            (not isinstance(key, str) or not key.startswith("_sa"))
            and (value is not None or not exclude_none)
            and key in allowed_keys
        ):
            encoded_key = jsonable_encoder(
                key,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_none=exclude_none,
            )
            encoded_value = jsonable_encoder(
                value,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_none=exclude_none,
            )
            encoded_dict[encoded_key] = encoded_value
    return encoded_dict


def _dump_sequence(
    *,
    obj: Any,
    include: Optional[IncEx] = None,
    exclude: Optional[IncEx] = None,
    by_alias: bool = True,
    exclude_unset: bool = False,
    exclude_none: bool = False,
    exclude_defaults: bool = False,
) -> List[Any]:
    """
    Dump a sequence to a list, using the same parameters as jsonable_encoder
    """
    encoded_list = []
    for item in obj:
        encoded_list.append(
            jsonable_encoder(
                item,
                include=include,
                exclude=exclude,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
            ),
        )
    return encoded_list


def _dump_other(
    *,
    obj: Any,
    include: Optional[IncEx] = None,
    exclude: Optional[IncEx] = None,
    by_alias: bool = True,
    exclude_unset: bool = False,
    exclude_none: bool = False,
    exclude_defaults: bool = False,
) -> Any:
    """
    Dump an object to ah hashable object, using the same parameters as jsonable_encoder
    """
    try:
        data = dict(obj)
    except Exception as e:
        errors: List[Exception] = [e]
        try:
            data = vars(obj)
        except Exception as e:
            errors.append(e)
            raise ValueError(errors) from e
    return jsonable_encoder(
        data,
        include=include,
        exclude=exclude,
        by_alias=by_alias,
        exclude_unset=exclude_unset,
        exclude_defaults=exclude_defaults,
        exclude_none=exclude_none,
    )


def iso_format(o: Union[datetime.date, datetime.time]) -> str:
    """
    ISO format for date and time
    """
    return o.isoformat()


def decimal_encoder(dec_value: Decimal) -> Union[int, float]:
    """
    Encodes a Decimal as int of there's no exponent, otherwise float

    This is useful when we use ConstrainedDecimal to represent Numeric(x,0)
    where an integer (but not int typed) is used. Encoding this as a float
    results in failed round-tripping between encode and parse.

    >>> decimal_encoder(Decimal("1.0"))
    1.0

    >>> decimal_encoder(Decimal("1"))
    1
    """
    if dec_value.as_tuple().exponent >= 0:  # type: ignore[operator]
        return int(dec_value)
    else:
        return float(dec_value)


# Encoders for types that are not JSON serializable
ENCODERS_BY_TYPE: Dict[Type[Any], Callable[[Any], Any]] = {
    bytes: lambda o: o.decode(),
    Color: str,
    datetime.date: iso_format,
    datetime.datetime: iso_format,
    datetime.time: iso_format,
    datetime.timedelta: lambda td: td.total_seconds(),
    Decimal: decimal_encoder,
    Enum: lambda o: o.value,
    frozenset: list,
    deque: list,
    GeneratorType: list,
    Path: str,
    Pattern: lambda o: o.pattern,
    SecretBytes: str,
    SecretStr: str,
    set: list,
    UUID: str,
}


# Generates a mapping of encoders to a tuple of classes that they can encode
def generate_encoders_by_class_tuples(
    type_encoder_map: Dict[Any, Callable[[Any], Any]],
) -> Dict[Callable[[Any], Any], Tuple[Any, ...]]:
    encoders: Dict[Callable[[Any], Any], Tuple[Any, ...]] = defaultdict(tuple)
    for type_, encoder in type_encoder_map.items():
        encoders[encoder] += (type_,)
    return encoders


# Mapping of encoders to a tuple of classes that they can encode
encoders_by_class_tuples = generate_encoders_by_class_tuples(ENCODERS_BY_TYPE)



"""
.. versionadded:: 20.7.0

Sometimes you want to confirm that your target data matches your
code's assumptions. With glom, you don't need a separate validation
step, you can do these checks inline with your glom spec, using
:class:`~glom.Match` and friends.
"""

import re
import sys
from pprint import pprint

from boltons.iterutils import is_iterable
from boltons.typeutils import make_sentinel

from .core import GlomError, glom, T, MODE, bbrepr, bbformat, format_invocation, Path, chain_child, Val, arg_val


_MISSING = make_sentinel('_MISSING')


# NOTE: it is important that MatchErrors be cheap to construct,
# because negative matches are part of normal control flow
# (e.g. often it is idiomatic to cascade from one possible match
# to the next and take the first one that works)
class MatchError(GlomError):
    """
    Raised when a :class:`Match` or :data:`M` check fails.

    >>> glom({123: 'a'}, Match({'id': int}))
    Traceback (most recent call last):
    ...
    MatchError: key 123 didn't match any of ['id']

    """
    def __init__(self, fmt, *args):
        super(MatchError, self).__init__(fmt, *args)

    def get_message(self):
        fmt, args = self.args[0], self.args[1:]
        return bbformat(fmt, *args)


class TypeMatchError(MatchError, TypeError):
    """:exc:`MatchError` subtype raised when a
    :class:`Match` fails a type check.

    >>> glom({'id': 'a'}, Match({'id': int}))
    Traceback (most recent call last):
    ...
    TypeMatchError: error raised while processing.
     Target-spec trace, with error detail (most recent last):
     - Target: {'id': 'a'}
     - Spec: Match({'id': <type 'int'>})
     - Spec: {'id': <type 'int'>}
     - Target: 'a'
     - Spec: int
    TypeMatchError: expected type int, not str
    """

    def __init__(self, actual, expected):
        super(TypeMatchError, self).__init__(
            "expected type {0.__name__}, not {1.__name__}", expected, actual)

    def __copy__(self):
        # __init__ args = (actual, expected)
        # self.args = (fmt_str, expected, actual)
        return TypeMatchError(self.args[2], self.args[1])


class Match(object):
    """glom's ``Match`` specifier type enables a new mode of glom usage:
    pattern matching.  In particular, this mode has been designed for
    nested data validation.

    Pattern specs are evaluated as follows:

      1. Spec instances are always evaluated first
      2. Types match instances of that type
      3. Instances of :class:`dict`, :class:`list`, :class:`tuple`,
         :class:`set`, and :class:`frozenset` are matched recursively
      4. Any other values are compared for equality to the target with
         ``==``

    By itself, this allows to assert that structures match certain
    patterns, and may be especially familiar to users of the `schema`_
    library.

    For example, let's load some data::

      >>> target = [
      ... {'id': 1, 'email': 'alice@example.com'},
      ... {'id': 2, 'email': 'bob@example.com'}]

    A :class:`Match` pattern can be used to ensure this data is in its expected form:

      >>> spec = Match([{'id': int, 'email': str}])

    This ``spec`` succinctly describes our data structure's pattern
    Specifically, a :class:`list` of :class:`dict` objects, each of
    which has exactly two keys, ``'id'`` and ``'email'``, whose values are
    an :class:`int` and :class:`str`, respectively. Now,
    :func:`~glom.glom` will ensure our ``target`` matches our pattern
    ``spec``:

      >>> result = glom(target, spec)
      >>> assert result == \\
      ... [{'id': 1, 'email': 'alice@example.com'}, {'id': 2, 'email': 'bob@example.com'}]

    With a more complex :class:`Match` spec, we can be more precise:

      >>> spec = Match([{'id': And(M > 0, int), 'email': Regex('[^@]+@[^@]+')}])

    :class:`~glom.And` allows multiple conditions to be applied.
    :class:`~glom.Regex` evaluates the regular expression against the
    target value under the ``'email'`` key.  In this case, we take a
    simple approach: an email has exactly one ``@``, with at least one
    character before and after.

    Finally, :attr:`~glom.M` is our stand-in for the current target
    we're matching against, allowing us to perform in-line comparisons
    using Python's native greater-than operator (as well as
    others). We apply our :class:`Match` pattern as before::

      >>> assert glom(target, spec) == \\
      ... [{'id': 1, 'email': 'alice@example.com'}, {'id': 2, 'email': 'bob@example.com'}]

    And as usual, upon a successful match, we get the matched result.

    .. note::

       For Python 3.6+ where dictionaries are ordered, keys in the target
       are matched against keys in the spec in their insertion order.

    .. _schema: https://github.com/keleshev/schema

    Args:
       spec: The glomspec representing the pattern to match data against.
       default: The default value to be returned if a match fails. If not
         set, a match failure will raise a :class:`MatchError`.

    """
    def __init__(self, spec, default=_MISSING):
        self.spec = spec
        self.default = default

    def glomit(self, target, scope):
        scope[MODE] = _glom_match
        try:
            ret = scope[glom](target, self.spec, scope)
        except GlomError:
            if self.default is _MISSING:
                raise
            ret = arg_val(target, self.default, scope)
        return ret

    def verify(self, target):
        """A convenience function a :class:`Match` instance which returns the
        matched value when *target* matches, or raises a
        :exc:`MatchError` when it does not.

        Args:
          target: Target value or data structure to match against.

        Raises:
          glom.MatchError

        """
        return glom(target, self)

    def matches(self, target):
        """A convenience method on a :class:`Match` instance, returns
        ``True`` if the *target* matches, ``False`` if not.

        >>> Match(int).matches(-1.0)
        False

        Args:
           target: Target value or data structure to match against.
        """
        try:
            glom(target, self)
        except GlomError:
            return False
        return True

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, bbrepr(self.spec))


_RE_FULLMATCH = getattr(re, "fullmatch", None)
_RE_VALID_FUNCS = set((_RE_FULLMATCH, None, re.search, re.match))
_RE_FUNC_ERROR = ValueError("'func' must be one of %s" % (", ".join(
    sorted(e and e.__name__ or "None" for e in _RE_VALID_FUNCS))))

_RE_TYPES = ()
try:   re.match(u"", u"")
except Exception: pass  # pragma: no cover
else:  _RE_TYPES += (type(u""),)
try:   re.match(b"", b"")
except Exception: pass  # pragma: no cover
else:  _RE_TYPES += (type(b""),)


class Regex(object):
    """
    checks that target is a string which matches the passed regex pattern

    raises MatchError if there isn't a match; returns Target if match

    variables captures in regex are added to the scope so they can
    be used by downstream processes
    """
    __slots__ = ('flags', 'func', 'match_func', 'pattern')

    def __init__(self, pattern, flags=0, func=None):
        if func not in _RE_VALID_FUNCS:
            raise _RE_FUNC_ERROR
        regex = re.compile(pattern, flags)
        if func is re.match:
            match_func = regex.match
        elif func is re.search:
            match_func = regex.search
        else:
            if _RE_FULLMATCH:
                match_func = regex.fullmatch
            else:
                regex = re.compile(r"(?:{})\Z".format(pattern), flags)
                match_func = regex.match
        self.flags, self.func = flags, func
        self.match_func, self.pattern = match_func, pattern

    def glomit(self, target, scope):
        if type(target) not in _RE_TYPES:
            raise MatchError(
                "{0!r} not valid as a Regex target -- expected {1!r}", type(target), _RE_TYPES)
        match = self.match_func(target)
        if not match:
            raise MatchError("target did not match pattern {0!r}", self.pattern)
        scope.update(match.groupdict())
        return target

    def __repr__(self):
        args = '(' + bbrepr(self.pattern)
        if self.flags:
            args += ', flags=' + bbrepr(self.flags)
        if self.func is not None:
            args += ', func=' + self.func.__name__
        args += ')'
        return self.__class__.__name__ + args


#TODO: combine this with other functionality elsewhere?
def _bool_child_repr(child):
    if child is M:
        return repr(child)
    elif isinstance(child, _MExpr):
        return "(" + bbrepr(child) + ")"
    return bbrepr(child)


class _Bool(object):
    def __init__(self, *children, **kw):
        self.children = children
        if not children:
            raise ValueError("need at least one operand for {}".format(
                self.__class__.__name__))
        self.default = kw.pop('default', _MISSING)
        if kw:
            raise TypeError('got unexpected kwargs: %r' % list(kw.keys()))

    def __and__(self, other):
        return And(self, other)

    def __or__(self, other):
        return Or(self, other)

    def __invert__(self):
        return Not(self)

    def glomit(self, target, scope):
        try:
            return self._glomit(target, scope)
        except GlomError:
            if self.default is not _MISSING:
                return arg_val(target, self.default, scope)
            raise

    def _m_repr(self):
        """should this Or() repr as M |?"""
        # only used by And() and Or(), not Not(), so len(children) >= 1
        if isinstance(self.children[0], (_MType, _MExpr)):
            return True
        if type(self.children[0]) in (And, Or, Not):
            return self.children[0]._m_repr()
        return False

    def __repr__(self):
        child_reprs = [_bool_child_repr(c) for c in self.children]
        if self._m_repr() and self.default is _MISSING:
            return " {} ".format(self.OP).join(child_reprs)
        if self.default is not _MISSING:
            child_reprs.append("default=" + repr(self.default))
        return self.__class__.__name__ + "(" + ", ".join(child_reprs) + ")"


class And(_Bool):
    """
    Applies child specs one after the other to the target; if none of the
    specs raises `GlomError`, returns the last result.
    """
    OP = "&"
    __slots__ = ('children',)

    def _glomit(self, target, scope):
        # all children must match without exception
        result = target  # so that And() == True, similar to all([]) == True
        for child in self.children:
            result = scope[glom](target, child, scope)
        return result

    def __and__(self, other):
        # reduce number of layers of spec
        return And(*(self.children + (other,)))


class Or(_Bool):
    """
    Tries to apply the first child spec to the target, and return the result.
    If `GlomError` is raised, try the next child spec until there are no
    all child specs have been tried, then raise `MatchError`.
    """
    OP = "|"
    __slots__ = ('children',)

    def _glomit(self, target, scope):
        for child in self.children[:-1]:
            try:  # one child must match without exception
                return scope[glom](target, child, scope)
            except GlomError:
                pass
        return scope[glom](target, self.children[-1], scope)

    def __or__(self, other):
        # reduce number of layers of spec
        return Or(*(self.children + (other,)))


class Not(_Bool):
    """
    Inverts the *child*. Child spec will be expected to raise
    :exc:`GlomError` (or subtype), in which case the target will be returned.

    If the child spec does not raise :exc:`GlomError`, :exc:`MatchError`
    will be raised.
    """
    __slots__ = ('child',)

    def __init__(self, child):
        self.child = child

    def glomit(self, target, scope):
        try:  # one child must match without exception
            scope[glom](target, self.child, scope)
        except GlomError:
            return target
        else:
            raise GlomError("child shouldn't have passed", self.child)

    def _m_repr(self):
        if isinstance(self.child, (_MType, _MExpr)):
            return True
        if type(self.child) not in (And, Or, Not):
            return False
        return self.child._m_repr()

    def __repr__(self):
        if self.child is M:
            return '~M'
        if self._m_repr():  # is in M repr
            return "~(" + bbrepr(self.child) + ")"
        return "Not(" + bbrepr(self.child) + ")"


_M_OP_MAP = {'=': '==', '!': '!=', 'g': '>=', 'l': '<='}


class _MSubspec(object):
    """used by MType.__call__ to wrap a sub-spec for comparison"""
    __slots__ = ('spec')

    def __init__(self, spec):
        self.spec = spec

    def __eq__(self, other):
        return _MExpr(self, '=', other)

    def __ne__(self, other):
        return _MExpr(self, '!', other)

    def __gt__(self, other):
        return _MExpr(self, '>', other)

    def __lt__(self, other):
        return _MExpr(self, '<', other)

    def __ge__(self, other):
        return _MExpr(self, 'g', other)

    def __le__(self, other):
        return _MExpr(self, 'l', other)

    def __repr__(self):
        return 'M(%s)' % (bbrepr(self.spec),)

    def glomit(self, target, scope):
        match = scope[glom](target, self.spec, scope)
        if match:
            return target
        raise MatchError('expected truthy value from {0!r}, got {1!r}', self.spec, match)


class _MExpr(object):
    __slots__ = ('lhs', 'op', 'rhs')

    def __init__(self, lhs, op, rhs):
        self.lhs, self.op, self.rhs = lhs, op, rhs

    def __and__(self, other):
        return And(self, other)

    __rand__ = __and__

    def __or__(self, other):
        return Or(self, other)

    def __invert__(self):
        return Not(self)

    def glomit(self, target, scope):
        lhs, op, rhs = self.lhs, self.op, self.rhs
        if lhs is M:
            lhs = target
        if rhs is M:
            rhs = target
        if type(lhs) is _MSubspec:
            lhs = scope[glom](target, lhs.spec, scope)
        if type(rhs) is _MSubspec:
            rhs = scope[glom](target, rhs.spec, scope)
        matched = (
            (op == '=' and lhs == rhs) or
            (op == '!' and lhs != rhs) or
            (op == '>' and lhs > rhs) or
            (op == '<' and lhs < rhs) or
            (op == 'g' and lhs >= rhs) or
            (op == 'l' and lhs <= rhs)
        )
        if matched:
            return target
        raise MatchError("{0!r} {1} {2!r}", lhs, _M_OP_MAP.get(op, op), rhs)

    def __repr__(self):
        op = _M_OP_MAP.get(self.op, self.op)
        return "{!r} {} {!r}".format(self.lhs, op, self.rhs)


class _MType(object):
    """:attr:`~glom.M` is similar to :attr:`~glom.T`, a stand-in for the
    current target, but where :attr:`~glom.T` allows for attribute and
    key access and method calls, :attr:`~glom.M` allows for comparison
    operators.

    If a comparison succeeds, the target is returned unchanged.
    If a comparison fails, :class:`~glom.MatchError` is thrown.

    Some examples:

    >>> glom(1, M > 0)
    1
    >>> glom(0, M == 0)
    0
    >>> glom('a', M != 'b') == 'a'
    True

    :attr:`~glom.M` by itself evaluates the current target for truthiness.
    For example, `M | Val(None)` is a simple idiom for normalizing all falsey values to None:

    >>> from glom import Val
    >>> glom([0, False, "", None], [M | Val(None)])
    [None, None, None, None]

    For convenience, ``&`` and ``|`` operators are overloaded to
    construct :attr:`~glom.And` and :attr:`~glom.Or` instances.

    >>> glom(1.0, (M > 0) & float)
    1.0

    .. note::

       Python's operator overloading may make for concise code,
       but it has its limits.

       Because bitwise operators (``&`` and ``|``) have higher precedence
       than comparison operators (``>``, ``<``, etc.), expressions must
       be parenthesized.

       >>> M > 0 & float
       Traceback (most recent call last):
       ...
       TypeError: unsupported operand type(s) for &: 'int' and 'type'

       Similarly, because of special handling around ternary
       comparisons (``1 < M < 5``) are implemented via
       short-circuiting evaluation, they also cannot be captured by
       :data:`M`.

    """
    __slots__ = ()

    def __call__(self, spec):
        """wrap a sub-spec in order to apply comparison operators to the result"""
        if not isinstance(spec, type(T)):
            # TODO: open this up for other specs so we can do other
            # checks, like function calls
            raise TypeError("M() only accepts T-style specs, not %s" % type(spec).__name__)
        return _MSubspec(spec)

    def __eq__(self, other):
        return _MExpr(self, '=', other)

    def __ne__(self, other):
        return _MExpr(self, '!', other)

    def __gt__(self, other):
        return _MExpr(self, '>', other)

    def __lt__(self, other):
        return _MExpr(self, '<', other)

    def __ge__(self, other):
        return _MExpr(self, 'g', other)

    def __le__(self, other):
        return _MExpr(self, 'l', other)

    def __and__(self, other):
        return And(self, other)

    __rand__ = __and__

    def __or__(self, other):
        return Or(self, other)

    def __invert__(self):
        return Not(self)

    def __repr__(self):
        return "M"

    def glomit(self, target, spec):
        if target:
            return target
        raise MatchError("{0!r} not truthy", target)


M = _MType()



class Optional(object):
    """Used as a :class:`dict` key in a :class:`~glom.Match()` spec,
    marks that a value match key which would otherwise be required is
    optional and should not raise :exc:`~glom.MatchError` even if no
    keys match.

    For example::

      >>> spec = Match({Optional("name"): str})
      >>> glom({"name": "alice"}, spec)
      {'name': 'alice'}
      >>> glom({}, spec)
      {}
      >>> spec = Match({Optional("name", default=""): str})
      >>> glom({}, spec)
      {'name': ''}

    """
    __slots__ = ('key', 'default')

    def __init__(self, key, default=_MISSING):
        if type(key) in (Required, Optional):
            raise TypeError("double wrapping of Optional")
        hash(key)  # ensure is a valid key
        if _precedence(key) != 0:
            raise ValueError("Optional() keys must be == match constants, not {!r}".format(key))
        self.key, self.default = key, default

    def glomit(self, target, scope):
        if target != self.key:
            raise MatchError("target {0} != spec {1}", target, self.key)
        return target

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, bbrepr(self.key))


class Required(object):
    """Used as a :class:`dict` key in :class:`~glom.Match()` mode, marks
    that a key which might otherwise not be required should raise
    :exc:`~glom.MatchError` if the key in the target does not match.

    For example::

      >>> spec = Match({object: object})

    This spec will match any dict, because :class:`object` is the base
    type of every object::

      >>> glom({}, spec)
      {}

    ``{}`` will also match because match mode does not require at
    least one match by default. If we want to require that a key
    matches, we can use :class:`~glom.Required`::

      >>> spec = Match({Required(object): object})
      >>> glom({}, spec)
      Traceback (most recent call last):
      ...
      MatchError: error raised while processing.
       Target-spec trace, with error detail (most recent last):
       - Target: {}
       - Spec: Match({Required(object): <type 'object'>})
       - Spec: {Required(object): <type 'object'>}
      MatchError: target missing expected keys Required(object)

    Now our spec requires at least one key of any type. You can refine
    the spec by putting more specific subpatterns inside of
    :class:`~glom.Required`.

    """
    __slots__ = ('key',)

    def __init__(self, key):
        if type(key) in (Required, Optional):
            raise TypeError("double wrapping of Required")
        hash(key)  # ensure is a valid key
        if _precedence(key) == 0:
            raise ValueError("== match constants are already required: " + bbrepr(key))
        self.key = key

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, bbrepr(self.key))


def _precedence(match):
    """
    in a dict spec, target-keys may match many
    spec-keys (e.g. 1 will match int, M > 0, and 1);
    therefore we need a precedence for which order to try
    keys in; higher = later
    """
    if type(match) in (Required, Optional):
        match = match.key
    if type(match) in (tuple, frozenset):
        if not match:
            return 0
        return max([_precedence(item) for item in match])
    if isinstance(match, type):
        return 2
    if hasattr(match, "glomit"):
        return 1
    return 0  # == match


def _handle_dict(target, spec, scope):
    if not isinstance(target, dict):
        raise TypeMatchError(type(target), dict)
    spec_keys = spec  # cheating a little bit here, list-vs-dict, but saves an object copy sometimes

    required = {
        key for key in spec_keys
        if _precedence(key) == 0 and type(key) is not Optional
        or type(key) is Required}
    defaults = {  # pre-load result with defaults
        key.key: key.default for key in spec_keys
        if type(key) is Optional and key.default is not _MISSING}
    result = {}
    for key, val in target.items():
        for maybe_spec_key in spec_keys:
            # handle Required as a special case here rather than letting it be a stand-alone spec
            if type(maybe_spec_key) is Required:
                spec_key = maybe_spec_key.key
            else:
                spec_key = maybe_spec_key
            try:
                key = scope[glom](key, spec_key, scope)
            except GlomError:
                pass
            else:
                result[key] = scope[glom](val, spec[maybe_spec_key], chain_child(scope))
                required.discard(maybe_spec_key)
                break
        else:
            raise MatchError("key {0!r} didn't match any of {1!r}", key, spec_keys)
    for key in set(defaults) - set(result):
        result[key] = arg_val(target, defaults[key], scope)
    if required:
        raise MatchError("target missing expected keys: {0}", ', '.join([bbrepr(r) for r in required]))
    return result


def _glom_match(target, spec, scope):
    if isinstance(spec, type):
        if not isinstance(target, spec):
            raise TypeMatchError(type(target), spec)
    elif isinstance(spec, dict):
        return _handle_dict(target, spec, scope)
    elif isinstance(spec, (list, set, frozenset)):
        if not isinstance(target, type(spec)):
            raise TypeMatchError(type(target), type(spec))
        result = []
        for item in target:
            for child in spec:
                try:
                    result.append(scope[glom](item, child, scope))
                    break
                except GlomError as e:
                    last_error = e
            else:  # did not break, something went wrong
                if target and not spec:
                    raise MatchError(
                        "{0!r} does not match empty {1}", target, type(spec).__name__)
                # NOTE: unless error happens above, break will skip else branch
                # so last_error will have been assigned
                raise last_error
        if type(spec) is not list:
            return type(spec)(result)
        return result
    elif isinstance(spec, tuple):
        if not isinstance(target, tuple):
            raise TypeMatchError(type(target), tuple)
        if len(target) != len(spec):
            raise MatchError("{0!r} does not match {1!r}", target, spec)
        result = []
        for sub_target, sub_spec in zip(target, spec):
            result.append(scope[glom](sub_target, sub_spec, scope))
        return tuple(result)
    elif callable(spec):
        try:
            if spec(target):
                return target
        except Exception as e:
            raise MatchError(
                "{0}({1!r}) did not validate (got exception {2!r})", spec.__name__, target, e)
        raise MatchError(
            "{0}({1!r}) did not validate (non truthy return)", spec.__name__, target)
    elif target != spec:
        raise MatchError("{0!r} does not match {1!r}", target, spec)
    return target


class Switch(object):
    r"""The :class:`Switch` specifier type routes data processing based on
    matching keys, much like the classic switch statement.

    Here is a spec which differentiates between lowercase English
    vowel and consonant characters:

      >>> switch_spec = Match(Switch([(Or('a', 'e', 'i', 'o', 'u'), Val('vowel')),
      ...                             (And(str, M, M(T[2:]) == ''), Val('consonant'))]))

    The constructor accepts a :class:`dict` of ``{keyspec: valspec}``
    or a list of items, ``[(keyspec, valspec)]``. Keys are tried
    against the current target in order. If a keyspec raises
    :class:`GlomError`, the next keyspec is tried.  Once a keyspec
    succeeds, the corresponding valspec is evaluated and returned.
    Let's try it out:

      >>> glom('a', switch_spec)
      'vowel'
      >>> glom('z', switch_spec)
      'consonant'

    If no keyspec succeeds, a :class:`MatchError` is raised. Our spec
    only works on characters (strings of length 1). Let's try a
    non-character, the integer ``3``:

      >>> glom(3, switch_spec)
      Traceback (most recent call last):
      ...
      glom.matching.MatchError: error raised while processing, details below.
       Target-spec trace (most recent last):
       - Target: 3
       - Spec: Match(Switch([(Or('a', 'e', 'i', 'o', 'u'), Val('vowel')), (And(str, M, (M(T[2:]) == '')), Val('...
       + Spec: Switch([(Or('a', 'e', 'i', 'o', 'u'), Val('vowel')), (And(str, M, (M(T[2:]) == '')), Val('conson...
       |\ Spec: Or('a', 'e', 'i', 'o', 'u')
       ||\ Spec: 'a'
       ||X glom.matching.MatchError: 3 does not match 'a'
       ||\ Spec: 'e'
       ||X glom.matching.MatchError: 3 does not match 'e'
       ||\ Spec: 'i'
       ||X glom.matching.MatchError: 3 does not match 'i'
       ||\ Spec: 'o'
       ||X glom.matching.MatchError: 3 does not match 'o'
       ||\ Spec: 'u'
       ||X glom.matching.MatchError: 3 does not match 'u'
       |X glom.matching.MatchError: 3 does not match 'u'
       |\ Spec: And(str, M, (M(T[2:]) == ''))
       || Spec: str
       |X glom.matching.TypeMatchError: expected type str, not int
      glom.matching.MatchError: no matches for target in Switch


    .. note::

       :class:`~glom.Switch` is one of several *branching* specifier
       types in glom. See ":ref:`branched-exceptions`" for details on
       interpreting its exception messages.

    A *default* value can be passed to the spec to be returned instead
    of raising a :class:`MatchError`.

    .. note::

      Switch implements control flow similar to the switch statement
      proposed in `PEP622 <https://www.python.org/dev/peps/pep-0622/>`_.

    """
    def __init__(self, cases, default=_MISSING):
        if type(cases) is dict:
            cases = list(cases.items())
        if type(cases) is not list:
            raise TypeError(
                "expected cases argument to be of format {{keyspec: valspec}}"
                " or [(keyspec, valspec)] not: {}".format(type(cases)))
        self.cases = cases
        # glom.match(cases, Or([(object, object)], dict))
        # start dogfooding ^
        self.default = default
        if not cases:
            raise ValueError('expected at least one case in %s, got: %r'
                             % (self.__class__.__name__, self.cases))
        return


    def glomit(self, target, scope):
        for keyspec, valspec in self.cases:
            try:
                scope[glom](target, keyspec, scope)
            except GlomError as ge:
                continue
            return scope[glom](target, valspec, chain_child(scope))
        if self.default is not _MISSING:
            return arg_val(target, self.default, scope)
        raise MatchError("no matches for target in %s"  % self.__class__.__name__)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, bbrepr(self.cases))


RAISE = make_sentinel('RAISE')  # flag object for "raise on check failure"


class Check(object):
    """Check objects are used to make assertions about the target data,
    and either pass through the data or raise exceptions if there is a
    problem.

    If any check condition fails, a :class:`~glom.CheckError` is raised.

    Args:

       spec: a sub-spec to extract the data to which other assertions will
          be checked (defaults to applying checks to the target itself)
       type: a type or sequence of types to be checked for exact match
       equal_to: a value to be checked for equality match ("==")
       validate: a callable or list of callables, each representing a
          check condition. If one or more return False or raise an
          exception, the Check will fail.
       instance_of: a type or sequence of types to be checked with isinstance()
       one_of: an iterable of values, any of which can match the target ("in")
       default: an optional default value to replace the value when the check fails
                (if default is not specified, GlomCheckError will be raised)

    Aside from *spec*, all arguments are keyword arguments. Each
    argument, except for *default*, represent a check
    condition. Multiple checks can be passed, and if all check
    conditions are left unset, Check defaults to performing a basic
    truthy check on the value.

    """
    # TODO: the next level of Check would be to play with the Scope to
    # allow checking to continue across the same level of
    # dictionary. Basically, collect as many errors as possible before
    # raising the unified CheckError.
    def __init__(self, spec=T, **kwargs):
        self.spec = spec
        self._orig_kwargs = dict(kwargs)
        self.default = kwargs.pop('default', RAISE)

        def _get_arg_val(name, cond, func, val, can_be_empty=True):
            if val is _MISSING:
                return ()
            if not is_iterable(val):
                val = (val,)
            elif not val and not can_be_empty:
                raise ValueError('expected %r argument to contain at least one value,'
                                 ' not: %r' % (name, val))
            for v in val:
                if not func(v):
                    raise ValueError('expected %r argument to be %s, not: %r'
                                     % (name, cond, v))
            return val

        # if there are other common validation functions, maybe a
        # small set of special strings would work as valid arguments
        # to validate, too.
        def truthy(val):
            return bool(val)

        validate = kwargs.pop('validate', _MISSING if kwargs else truthy)
        type_arg = kwargs.pop('type', _MISSING)
        instance_of = kwargs.pop('instance_of', _MISSING)
        equal_to = kwargs.pop('equal_to', _MISSING)
        one_of = kwargs.pop('one_of', _MISSING)
        if kwargs:
            raise TypeError('unexpected keyword arguments: %r' % kwargs.keys())

        self.validators = _get_arg_val('validate', 'callable', callable, validate)
        self.instance_of = _get_arg_val('instance_of', 'a type',
                                        lambda x: isinstance(x, type), instance_of, False)
        self.types = _get_arg_val('type', 'a type',
                                  lambda x: isinstance(x, type), type_arg, False)

        if equal_to is not _MISSING:
            self.vals = (equal_to,)
            if one_of is not _MISSING:
                raise TypeError('expected "one_of" argument to be unset when'
                                ' "equal_to" argument is passed')
        elif one_of is not _MISSING:
            if not is_iterable(one_of):
                raise ValueError('expected "one_of" argument to be iterable'
                                 ' , not: %r' % one_of)
            if not one_of:
                raise ValueError('expected "one_of" to contain at least'
                                 ' one value, not: %r' % (one_of,))
            self.vals = one_of
        else:
            self.vals = ()
        return

    class _ValidationError(Exception):
        "for internal use inside of Check only"
        pass

    def glomit(self, target, scope):
        ret = target
        errs = []
        if self.spec is not T:
            target = scope[glom](target, self.spec, scope)
        if self.types and type(target) not in self.types:
            if self.default is not RAISE:
                return arg_val(target, self.default, scope)
            errs.append('expected type to be %r, found type %r' %
                        (self.types[0].__name__ if len(self.types) == 1
                         else tuple([t.__name__ for t in self.types]),
                         type(target).__name__))

        if self.vals and target not in self.vals:
            if self.default is not RAISE:
                return arg_val(target, self.default, scope)
            if len(self.vals) == 1:
                errs.append("expected {}, found {}".format(self.vals[0], target))
            else:
                errs.append('expected one of {}, found {}'.format(self.vals, target))

        if self.validators:
            for i, validator in enumerate(self.validators):
                try:
                    res = validator(target)
                    if res is False:
                        raise self._ValidationError
                except Exception as e:
                    msg = ('expected %r check to validate target'
                           % getattr(validator, '__name__', None) or ('#%s' % i))
                    if type(e) is self._ValidationError:
                        if self.default is not RAISE:
                            return self.default
                    else:
                        msg += ' (got exception: %r)' % e
                    errs.append(msg)

        if self.instance_of and not isinstance(target, self.instance_of):
            # TODO: can these early returns be done without so much copy-paste?
            # (early return to avoid potentially expensive or even error-causeing
            # string formats)
            if self.default is not RAISE:
                return arg_val(target, self.default, scope)
            errs.append('expected instance of %r, found instance of %r' %
                        (self.instance_of[0].__name__ if len(self.instance_of) == 1
                         else tuple([t.__name__ for t in self.instance_of]),
                         type(target).__name__))

        if errs:
            raise CheckError(errs, self, scope[Path])
        return ret

    def __repr__(self):
        cn = self.__class__.__name__
        posargs = (self.spec,) if self.spec is not T else ()
        return format_invocation(cn, posargs, self._orig_kwargs, repr=bbrepr)


class CheckError(GlomError):
    """This :exc:`GlomError` subtype is raised when target data fails to
    pass a :class:`Check`'s specified validation.

    An uncaught ``CheckError`` looks like this::

       >>> target = {'a': {'b': 'c'}}
       >>> glom(target, {'b': ('a.b', Check(type=int))})
       Traceback (most recent call last):
       ...
       CheckError: target at path ['a.b'] failed check, got error: "expected type to be 'int', found type 'str'"

    If the ``Check`` contains more than one condition, there may be
    more than one error message. The string rendition of the
    ``CheckError`` will include all messages.

    You can also catch the ``CheckError`` and programmatically access
    messages through the ``msgs`` attribute on the ``CheckError``
    instance.

    """
    def __init__(self, msgs, check, path):
        self.msgs = msgs
        self.check_obj = check
        self.path = path

    def get_message(self):
        msg = 'target at path %s failed check,' % self.path
        if self.check_obj.spec is not T:
            msg += ' subtarget at %r' % (self.check_obj.spec,)
        if len(self.msgs) == 1:
            msg += ' got error: %r' % (self.msgs[0],)
        else:
            msg += ' got %s errors: %r' % (len(self.msgs), self.msgs)
        return msg

    def __repr__(self):
        cn = self.__class__.__name__
        return '%s(%r, %r, %r)' % (cn, self.msgs, self.check_obj, self.path)
